# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from datetime import timedelta
from typing import ClassVar, Optional, TYPE_CHECKING

from .api import MetricEvents
from ..model.core import Party

if TYPE_CHECKING:
    from prometheus_client import Gauge
    from prometheus_client.registry import CollectorRegistry


class PrometheusMetricEvents(MetricEvents):

    _DEFAULT = None  # type: ClassVar[PrometheusMetricEvents]

    @classmethod
    def default(cls):
        if PrometheusMetricEvents._DEFAULT is not None:
            return PrometheusMetricEvents._DEFAULT
        else:
            from prometheus_client import REGISTRY
            PrometheusMetricEvents._DEFAULT = PrometheusMetricEvents(REGISTRY)
            return PrometheusMetricEvents._DEFAULT

    def __init__(self, registry: 'Optional[CollectorRegistry]' = None):
        from prometheus_client import Gauge, REGISTRY

        if registry is None:
            registry = REGISTRY

        self._loop_responsiveness_seconds = Gauge(
            'dazl_network_loop_responsiveness_seconds',
            'Number of seconds to immediately schedule and execute a no-op on the event loop',
            registry=registry)
        self._offset = Gauge(
            'dazl_party_offset',
            'current read offset',
            ['party'],
            registry=registry)

    def loop_responsiveness(self, responsiveness: 'timedelta') -> None:
        self._loop_responsiveness_seconds.set(responsiveness.total_seconds())

    def party_offset(self, party: 'Party', offset: str) -> None:
        self._offset.labels(party).set(offset)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Support for the gRPC-based Ledger API.
"""
from asyncio import gather, get_event_loop
from datetime import datetime
from dataclasses import replace
from threading import Thread, Event
from typing import Awaitable, Iterable, Optional, Sequence

# noinspection PyPackageRequirements
from grpc import Channel, secure_channel, insecure_channel, ssl_channel_credentials, RpcError

from ... import LOG
from .._base import LedgerClient, _LedgerConnection, _LedgerConnectionContext
from .grpc_time import maybe_grpc_time_stream
from .pb_parse_event import serialize_acs_request, serialize_event_id_request, \
    serialize_transactions_request, to_acs_events, to_transaction_events, \
    BaseEventDeserializationContext
from .pb_parse_metadata import parse_daml_metadata_pb, parse_archive_payload, find_dependencies
from ...model.core import Party
from ...model.ledger import LedgerMetadata, StaticTimeModel, RealTimeModel
from ...model.network import HTTPConnectionSettings
from ...model.reading import BaseEvent, TransactionFilter, ActiveContractSetEvent
from ...model.types_store import PackageStore, PackageProvider
from ...model.writing import CommandPayload
from ...util.io import read_file_bytes
from ...util.prim_types import to_datetime
from ...util.typing import safe_cast


class GRPCv1LedgerClient(LedgerClient):
    def __init__(self, connection: 'GRPCv1Connection', ledger: LedgerMetadata, party: Party):
        self.connection = safe_cast(GRPCv1Connection, connection)
        self.ledger = safe_cast(LedgerMetadata, ledger)
        self.party = Party(safe_cast(str, party))

    def commands(self, commands: CommandPayload) -> None:
        serializer = self.ledger.serializer
        request = serializer.serialize_command_request(commands)
        return self.connection.context.run_in_background(
            lambda: self.connection.command_service.SubmitAndWait(request))

    async def active_contracts(self) -> 'Sequence[BaseEvent]':
        request = serialize_acs_request(self.ledger.ledger_id, self.party)
        print(request)
        context = BaseEventDeserializationContext(
            None, self.ledger.store, self.party, self.ledger.ledger_id)
        acs_stream = await self.connection.context.run_in_background(
            lambda: self.connection.active_contract_set_service.GetActiveContracts(request))
        acs_events = to_acs_events(context, acs_stream)
        return await self._augmented_acs_events(acs_events)

    def events(self, transaction_filter: TransactionFilter) \
            -> Awaitable[Sequence[BaseEvent]]:
        results_future = get_event_loop().create_future()
        request = serialize_transactions_request(transaction_filter, self.party)

        context = BaseEventDeserializationContext(
            None, self.ledger.store, self.party, transaction_filter.ledger_id)

        def on_events_done(fut):
            try:
                tx_stream, tt_stream = fut.result()
                events = to_transaction_events(
                    context, tx_stream, tt_stream, transaction_filter.destination_offset)
            except Exception as ex:
                LOG.exception('Failed to parse data coming back from an event!')
                results_future.set_exception(ex)
                return
            results_future.set_result(events)

        ts_future = self.connection.context.run_in_background(
            lambda: self.connection.transaction_service.GetTransactions(request))
        tst_future = self.connection.context.run_in_background(
            lambda: self.connection.transaction_service.GetTransactionTrees(request))
        t_fut = gather(ts_future, tst_future)
        t_fut.add_done_callback(on_events_done)

        return results_future

    async def events_end(self) -> str:
        from . import model as G
        request = G.GetLedgerEndRequest(ledger_id=self.ledger.ledger_id)
        return await self.connection.context.run_in_background(
            lambda: self.connection.transaction_service.GetLedgerEnd(request).offset.absolute)

    async def _augmented_acs_events(self, acs_events: 'Sequence[ActiveContractSetEvent]')  \
            -> 'Sequence[ActiveContractSetEvent]':
        requests = [serialize_event_id_request(self.ledger.ledger_id, event_id, [self.party])
                    for event_id
                    in {e.event_id for acs_event in acs_events for e in acs_event.contract_events}]
        print(requests)
        responses = await gather(*(self.connection.context.run_in_background(
            lambda: self.connection.transaction_service.GetFlatTransactionByEventId(request)
        ) for request in requests))

        event_times = {}
        for response in responses:
            dt = to_datetime(response.transaction.effective_at)
            for event in response.transaction.events:
                if event.WhichOneof('event') == 'created':
                    event_times[event.created.event_id] = dt
                else:
                    print(event.archived.event_id)
        print(event_times)

        return [replace(acs_event, contract_events=[replace(contract_event, time=event_times.get(contract_event.event_id))
                                                    for contract_event in acs_event.contract_events])
                for acs_event in acs_events]


def grpc_set_time(connection: 'GRPCv1Connection', ledger_id: str, new_datetime: datetime) -> None:
    from . import model as G
    from .pb_ser_command import as_api_timestamp

    request = G.GetTimeRequest(ledger_id=ledger_id)
    response = connection.time_service.GetTime(request)
    ts = next(iter(response))

    request = G.SetTimeRequest(
        ledger_id=ledger_id,
        current_time=ts.current_time,
        new_time=as_api_timestamp(new_datetime))
    connection.time_service.SetTime(request)
    LOG.info('Time on the server changed by the local client to %s.', new_datetime)


def grpc_detect_ledger_id(stub: 'GRPCv1Connection') -> Optional[str]:
    """
    Return the ledger ID from the remote server when it becomes available, or ``None`` if the
    Ledger Identity API is not supported.
    """
    from . import model as G
    LOG.debug("Starting a monitor thread for stubs: %s", stub)
    try:
        response = stub.ledger_identity_service.GetLedgerIdentity(G.GetLedgerIdentityRequest())
    except RpcError as ex:
        if ex.details().startswith('Received http2 header with status'):
            LOG.info("No gRPC Ledger Identity Service found!")
            return None
        LOG.exception('An unexpected error occurred when trying to fetch the ledger identity.')
        raise

    return response.ledger_id


def grpc_main_thread(connection: 'GRPCv1Connection', ledger_id: str) -> Iterable[LedgerMetadata]:
    from . import model as G
    from .pb_ser_command import ProtobufSerializer

    store = PackageStore.empty()
    request = G.GetLedgerEndRequest(ledger_id=ledger_id)

    package_provider = GRPCPackageProvider(connection, ledger_id)

    grpc_package_sync(package_provider, store)

    time_iter = maybe_grpc_time_stream(connection.time_service, ledger_id)
    time_model = StaticTimeModel(next(time_iter)) if time_iter is not None else RealTimeModel()

    yield LedgerMetadata(
        ledger_id=ledger_id,
        store=store,
        time_model=time_model,
        serializer=ProtobufSerializer(store),
        protocol_version="v1")

    # oh man, another thread...
    def time_sync_thread():
        try:
            # keep the network open as long as we keep getting time updates
            for ts in time_iter:
                time_model.current_time = ts
            LOG.debug('The time stream has completed.')
        except:
            pass

    if time_iter is not None:
        time_thread = Thread(name='time-sync-thread', target=time_sync_thread, daemon=True)
        time_thread.start()

    LOG.debug('There is no time service, so we will merely wait for the underlying transaction '
              'stream to die.')
    while True:
        grpc_package_sync(package_provider, store)
        if connection._closed.wait(1):
            break

    LOG.debug('The gRPC monitor thread is now shutting down.')
    yield None


class GRPCPackageProvider(PackageProvider):
    def __init__(self, connection: 'GRPCv1Connection', ledger_id: str):
        self.connection = connection
        self.ledger_id = ledger_id

    def get_package_ids(self) -> 'Sequence[str]':
        from . import model as G
        request = G.ListPackagesRequest(ledger_id=self.ledger_id)
        response = self.connection.package_service.ListPackages(request)
        return response.package_ids

    def fetch_package(self, package_id: str) -> bytes:
        from . import model as G
        request = G.GetPackageRequest(ledger_id=self.ledger_id, package_id=package_id)
        package_info = self.connection.package_service.GetPackage(request)
        return package_info.archive_payload


def grpc_package_sync(package_provider: PackageProvider, store: 'PackageStore') -> 'None':
    all_package_ids = package_provider.get_package_ids()
    loaded_package_ids = [a.hash for a in store.archives()]

    metadatas_pb = {}
    for package_id in all_package_ids:
        if package_id not in loaded_package_ids:
            archive_payload = package_provider.fetch_package(package_id)
            metadatas_pb[package_id] = parse_archive_payload(archive_payload)

    metadatas_pb = find_dependencies(metadatas_pb, loaded_package_ids)
    for package_id, archive_payload in metadatas_pb.items():
        store.register_all(parse_daml_metadata_pb(package_id, archive_payload))


def grpc_create_channel(settings: HTTPConnectionSettings) -> Channel:
    target = f'{settings.host}:{settings.port}'
    options = [('grpc.max_send_message_length', -1),
               ('grpc.max_receive_message_length', -1)]

    if settings.oauth:
        # noinspection PyPackageRequirements
        from google.auth.transport.grpc import secure_authorized_channel
        # noinspection PyPackageRequirements
        from google.auth.transport.requests import Request as RefreshRequester
        # noinspection PyPackageRequirements
        from google.oauth2.credentials import Credentials as OAuthCredentials

        LOG.debug('Using a secure gRPC connection over OAuth:')

        credentials = OAuthCredentials(
            token=settings.oauth.token,
            refresh_token=settings.oauth.refresh_token,
            id_token=settings.oauth.id_token,
            token_uri=settings.oauth.token_uri,
            client_id=settings.oauth.client_id,
            client_secret=settings.oauth.client_secret)
        return secure_authorized_channel(credentials, RefreshRequester(), target, options=options)

    if settings.ssl_settings:
        cert_chain = read_file_bytes(settings.ssl_settings.cert_file)
        cert = read_file_bytes(settings.ssl_settings.cert_key_file)
        ca_cert = read_file_bytes(settings.ssl_settings.ca_file)

        LOG.debug('Using a secure gRPC connection:')
        LOG.debug('    target: %s', target)
        LOG.debug('    root_certificates: contents of %s', settings.ssl_settings.ca_file)
        LOG.debug('    private_key: contents of %s', settings.ssl_settings.cert_key_file)
        LOG.debug('    certificate_chain: contents of %s', settings.ssl_settings.cert_file)

        credentials = ssl_channel_credentials(
            root_certificates=ca_cert,
            private_key=cert,
            certificate_chain=cert_chain)
        return secure_channel(target, credentials, options)
    else:
        LOG.debug('Using an insecure gRPC connection...')
        return insecure_channel(target, options)


class GRPCv1Connection(_LedgerConnection):
    def __init__(self,
                 context: _LedgerConnectionContext,
                 settings: HTTPConnectionSettings,
                 context_path: Optional[str]):
        super(GRPCv1Connection, self).__init__(context, settings, context_path)
        from . import model as G

        self._closed = Event()
        self._channel = grpc_create_channel(settings)
        self.active_contract_set_service = G.ActiveContractsServiceStub(self._channel)
        self.command_service = G.CommandServiceStub(self._channel)
        self.transaction_service = G.TransactionServiceStub(self._channel)
        self.package_service = G.PackageServiceStub(self._channel)
        self.ledger_identity_service = G.LedgerIdentityServiceStub(self._channel)
        self.time_service = G.TimeServiceStub(self._channel)

    def close(self):
        try:
            self._closed.set()
            self._channel.close()
        except:
            LOG.warning('An exception was thrown when trying to close down connections.')
        finally:
            super().close()




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from .expr_eval import ExpressionEvaluator
from .type_eval import TypeComputer



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Core types
----------

The :mod:`dazl.model.core` module contains classes used on both the read-side and the write-side of
the Ledger API.

.. autoclass:: ContractId
   :members:
"""
import warnings
from dataclasses import dataclass
from typing import Any, Callable, Collection, Dict, List, NewType, Optional, Tuple, TypeVar, \
    Union, TYPE_CHECKING
from datetime import datetime

from ..util.enum import OrderedEnum

if TYPE_CHECKING:
    from .types import Type

T = TypeVar('T')


class ContractId:
    """
    There are two kinds of contract IDs: those that know the template type of the underlying
    contract instance and those that don't. Contract IDs that arise from event processing are always
    tagged with their type when they are read off the event stream. Contract IDs that are
    parameters to a template are not currently tagged with a template type.

    Instance attributes:

    .. attribute:: ContractId.contract_id

        A ``str`` reference to a contract.

    .. attribute:: ContractId.template_id

        An optional ``str`` template ID.

    Instance members:
    """
    __slots__ = ('contract_id', 'template_id')

    def __init__(self, contract_id: str, template_id: 'Union[None, str, Type]' = None):
        from .types import Type, UnresolvedTypeReference
        if not isinstance(contract_id, str):
            raise ValueError('contract_id must be a string')

        if template_id is None:
            warnings.warn('Untyped ContractId support will be removed with the removal of '
                          'the deprecated REST API.', DeprecationWarning, stacklevel=2)
        else:
            if isinstance(template_id, str):
                template_id = UnresolvedTypeReference(template_id)
            elif not isinstance(template_id, Type):
                raise ValueError(f'template_id must either be unspecified or a template identifier '
                                 f'(got {repr(template_id)})')

        self.contract_id = contract_id
        self.template_id = template_id

    def __str__(self):
        """
        Return the contract ID without a type adornment.
        """
        return self.contract_id

    def __repr__(self):
        if self.template_id is not None:
            return '<ContractId("{}<{}>")>'.format(self.contract_id, self.template_id)
        else:
            return '<ContractId("{}")>'.format(self.contract_id)

    def __eq__(self, other):
        """
        Returns whether this contract is the same as the other one. Template
        type is NOT considered in equality.
        """
        return isinstance(other, ContractId) and self.contract_id == other.contract_id

    def __format__(self, format_spec):
        return ('{:' + format_spec + 's}').format(self.contract_id)

    def __hash__(self):
        """
        Returns a hash of the ContractId (based on the value of ContractId).
        """
        return hash(self.contract_id)

    def exercise(self, choice_name, arguments=None):
        """
        Create an :class:`ExerciseCommand` that represents the result of exercising a choice on this
        contract with the specified choice.

        :param choice_name:
            The name of the choice to exercise.
        :param arguments:
            (optional) A ``dict`` of named values to send as parameters to the choice exercise.
        """
        from .writing import ExerciseCommand
        return ExerciseCommand(self, choice_name, arguments=arguments)

    def replace(self, contract_id=None, template_id=None):
        """
        Return a new :class:`ContractId` instance replacing specified fields with values.
        """
        return ContractId(
            contract_id if contract_id is not None else self.contract_id,
            template_id if template_id is not None else self.template_id)

    def for_json(self):
        """
        Return the JSON representation of this contract. This is currently just the contract ID
        string itself.
        """
        return self.contract_id


ContractData = Dict[str, Any]
ContractMatch = Union[None, Callable[[ContractData], bool], ContractData]
ContractsState = Dict[ContractId, ContractData]
ContractsHistoricalState = Dict[ContractId, Tuple[ContractData, bool]]
Party = NewType('Party', str)


class ContractContextualDataCollection(tuple):

    def __getitem__(self, index: Union[int, str, ContractId]):
        if index is None:
            raise ValueError('the index cannot be None')
        elif isinstance(index, int):
            return tuple.__getitem__(self, index)
        elif isinstance(index, str):
            for cxd in self:
                if cxd.cid.contract_id == index:
                    return cxd
            raise KeyError(index)
        elif isinstance(index, ContractId):
            for cxd in self:
                if cxd.cid == index:
                    return cxd
            raise KeyError(index)
        else:
            raise TypeError("cannot index into a ContractContextualDataCollection with {index!r}")


@dataclass(frozen=True)
class ContractContextualData:
    cid: ContractId
    cdata: 'Optional[ContractData]'
    effective_at: datetime
    archived_at: 'Optional[datetime]'
    active: bool


class RunLevel(OrderedEnum):
    RUN_FOREVER = 0
    RUN_UNTIL_IDLE = 1
    TERMINATE_GRACEFULLY = 2
    TERMINATE_IMMEDIATELY = 3
    STOPPED = 4


class DazlError(Exception):
    """
    Superclass of errors raised by dazl.
    """


class DazlWarning(Warning):
    """
    Superclass of warnings raised by dazl.
    """


class CommandTimeoutError(DazlError):
    """
    Raised when a corresponding event for a command was not seen in the appropriate time window.
    """


class ConfigurationError(DazlError):
    """
    Raised when a configuration error prevents a client from being started.

    .. attribute:: ConfigurationError.reasons

        A collection of reasons for a failure.
    """
    def __init__(self, reasons: 'Union[str, Collection[str]]'):
        if reasons is None:
            self.reasons = []  # type: List[str]
        elif isinstance(reasons, str):
            self.reasons = [reasons]
        else:
            self.reasons = reasons  # type: List[str]


class ConnectionClosedError(DazlError):
    """
    Raised when trying to do something that requires a connection after connection pools have been
    closed.
    """


class UnknownTemplateWarning(DazlWarning):
    """
    Raised when trying to do something with a template name that is unknown.
    """


class ProcessDiedException(DazlError):
    pass



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import asyncio
import copy
import itertools
import functools

from asyncio import Future
from collections import defaultdict
from functools import partial
from io import StringIO
from typing import Any, Callable, Tuple, Dict

from .. import LOG
from ..client import LedgerClientManager, ParticipantLedgerClient
from ..util.tools import flatten
from ..model.reading import TransactionStartEvent
from ..plugins import Plugin

from .appbuilder import Application
from .builtins import BUILTIN_AND
from .visitor_eval import generate_lambda
from .visitor_pretty import pretty_print
from .expr_base import TRIGGER_ON_INIT
from .expr_impl import and_, AppExpr
from .expr_util import get_distinct_templates, get_distinct_aliases


class ApplicationContext(Plugin):
    """
    Manages the lifetime of a running Application.
    """
    def __init__(self, application):
        self.suite_contexts = tuple(map(SuiteContext, application.suites))

    def install(self, manager):
        for suite_context in self.suite_contexts:
            suite_context.register_into(manager.client(suite_context.party))


class SuiteContext:
    """
    Manages the lifetime of a suite of programs associated with a single party.
    """
    def __init__(self, suite, loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self.suite = suite
        self.client = None

    @property
    def party(self):
        return self.suite.party

    def _init_ready(self, program):
        def _invoke_init_program(party, client):
            return generate_lambda(program.update)(dict())
        return _invoke_init_program

    def register_into(self, client: ParticipantLedgerClient):
        self.client = client

        for program in self.suite.programs:
            # init programs with no preconditions runs first
            if program.trigger == TRIGGER_ON_INIT and program.select is None:
                # these special programs run with the trivial initial state
                client.on_ready(self._init_ready(program))
            else:
                # these special programs are supposed to only when crossing a boundary of a
                # pre-state not matching and post-state matching
                # TODO: Proper implementation; this implementation is too aggressive
                template_refs = get_distinct_templates(program)
                if not template_refs:
                    LOG.error('A program somehow referenced no templates: %r', program)

                runner = Runner(self.loop, self.client, program, flatten(template_refs.values()))
                client.on_event('transaction-start', runner.start_transaction)
                for template_ref, aliases in template_refs.items():
                    client.on_event(template_ref, runner.state_updater(aliases))
                client.on_event('transaction-end', runner.end_transaction)
                client.on_done(runner.log_post_checks)

                # have the program force the application to stay alive if it is an `init` program
                client._keepalive.wait_for_future(runner.run(), keep_alive=program.trigger == TRIGGER_ON_INIT)


def register_application(client_manager: LedgerClientManager, application: Application):
    """
    Install listeners into clients managed by a `LedgerClientManager` that allow for
    starting an application.
    """
    client_manager.register(ApplicationContext(application))


def trivial_future(loop, result='TRIVIAL_FUTURE'):
    future = loop.create_future()
    future.set_result(result)
    return future


class _TransactionBuilder:
    __slots__ = ('bim', 'events')

    def __init__(self):
        self.bim = None
        self.events = []

    def build(self):
        return _Transaction(self.bim, self.events)


class _Transaction:
    __slots__ = ('bim', 'events')

    def __init__(self, bim, events):
        self.bim = bim
        self.events = tuple(events)


class ProgramMetrics:
    """
    Metrics on a running program.
    """
    __slots__ = ('trigger_count', 'update_count')

    def __init__(self):
        self.trigger_count = 0
        self.update_count = 0


class _ProgramState:
    """
    Tracks all state required by a running program.
    """
    def __init__(self, loop, client, program, state_predicate_fn, defined_aliases):
        self.loop = loop
        self.client = client
        self._program = program

        self._state_predicate_fn = state_predicate_fn
        self._update_fn = generate_lambda(program.update)
        self._state = {alias: dict() for alias in defined_aliases}
        self._current_command = None
        self.metrics = ProgramMetrics()
        self._error = None

    def apply_transaction(self, tx: _Transaction) -> bool:
        """
        Applies the state as necessary and submits the corresponding update to the server.

        :return:
            ``True`` generally; ``False`` if the state of the program dictates that it needs to
            terminate.
        """
        # TODO: Holy moly, what a hack.
        # This essentially stops `init`-based programs from being retriggered once their initial
        # entry criteria have been met.
        if self._program.trigger == TRIGGER_ON_INIT and self.metrics.update_count:
            self._current_command = None
            return False

        if self._error is not None:
            return False

        for alias, cid, cdata in tx.events:
            if cdata is not None:
                self._state[alias][cid] = cdata
            else:
                self._state[alias].pop(cid, None)

        self._attempt_submit()
        return True

    def _attempt_submit(self):
        if self._current_command is None or self._current_command.done():
            self._current_command = None
            self.metrics.trigger_count += 1
            update = self.compute_update()
            if update:
                # TODO: questionable...
                if self._program.trigger == TRIGGER_ON_INIT and self.metrics.update_count:
                    return

                self.metrics.update_count += 1

                self._current_command = self.client.submit(update)
                self._current_command.state_snapshot = copy.deepcopy(self._state)
                self._current_command.update_command = update
                self._current_command.add_done_callback(self._check_for_errors)

    def _check_for_errors(self, future):
        try:
            if future.exception() is not None:
                self._error = True
                with StringIO() as buf:
                    try:
                        buf.write('Failed to execute a program!\n')
                        self.write_state(snapshot=future.state_snapshot, buf=buf)
                        buf.write('\nProposed Update:\n')
                        buf.write(str(future.update_command))
                        buf.write('\nServer Error\n')
                        buf.write(repr(future.exception()))
                        LOG.error(buf.getvalue())
                    finally:
                        LOG.exception("Failed to log an exception!")
        except:
            LOG.exception('Error when trying to process an error result.')

    def compute_update(self):
        """
        Computes the fully-realized update block from the current state.
        """
        return compute_update(self._state, self._state_predicate_fn, self._update_fn)

    def write_state(self, snapshot=None, buf=None):
        if buf is None:
            buf = StringIO()
        if snapshot is None:
            snapshot = self._state

        buf.write('Program:\n')
        buf.write(pretty_print(self._program))
        buf.write('\nState:\n')
        for alias, data in snapshot.items():
            buf.write('  {} ({} item(s))\n'.format(alias, len(data)))
            for cid, cdata in data.items():
                buf.write('    ')
                buf.write(str(cid))
                buf.write(' ')
                buf.write(str(cdata))
                buf.write('\n')
        return buf

    def log_post_checks(self):
        """
        Log warnings based on operations that have (or more helpfully, have NOT) been performed.
        """
        if self.metrics.update_count == 0:
            # this is almost certainly a problem; there is no use for a program that never produces
            # an update
            LOG.warning('The program %r for party %r never produced an update!', self._program.name, self.client.party_name)
            LOG.warning('  update count: %r', self.metrics.update_count)
            LOG.warning('  trigger count: %r', self.metrics.trigger_count)
            with StringIO() as buf:
                self.write_state(buf=buf)
                LOG.warning(buf.getvalue())
        if self._current_command is not None and not self._current_command.done():
            LOG.warning('The program %r for party %r is stuck waiting for an update!', self._program.name, self.client.party_name)
            LOG.warning('  stuck command: %r', self._current_command)
            LOG.warning('  update count: %r', self.metrics.update_count)
            LOG.warning('  trigger count: %r', self.metrics.trigger_count)
            with StringIO() as buf:
                self.write_state(buf=buf)
                LOG.warning(buf.getvalue())

    def __repr__(self):
        return f'<_ProgramState(name={self.client.party_name})>'


# Filters that are applied on the full state tracked by a program.
#
# This is a function that takes one row of the full state of a program and returns whether the row
# should be considered part of the state.
StateFilter = Callable[[Dict[str, Tuple[Any, Any]]], bool]

# Filters that are applied before data is applied to the state tracked by a program over
# a subset of the data.
#
# This is a dictionary whose keys are aliases and values are functions that take a state, assumed
# to only contain one alias for the (cid, cdata) tuple and return whether the (cid, cdata) is valid
# to include in the state.
PreFilter = Dict[str, StateFilter]


def tautology(state):
    return True


def compute_filters(root_predicate: Any) -> Tuple[PreFilter, StateFilter]:
    if isinstance(root_predicate, AppExpr) and root_predicate.func == BUILTIN_AND:
        # We currently only compute pre-filters on top-level AND expressions that are functions
        # of exclusively one alias. All other expressions are pushed off to a post-filter,
        # which is more expensive.
        pre_filter_predicates = defaultdict(list)
        post_filter_predicates = []
        for child in root_predicate.args:
            aliases = get_distinct_aliases(child)
            if len(aliases) == 1:
                alias, = aliases
                pre_filter_predicates[alias].append(child)
            else:
                post_filter_predicates.append(child)

        pre_filters = {alias: generate_lambda(and_(*predicates))
                       for alias, predicates in pre_filter_predicates.items() if predicates}
        post_filter = generate_lambda(and_(*post_filter_predicates)) if post_filter_predicates else tautology
    else:
        # can't optimize this expression further
        pre_filters = dict()
        post_filter = generate_lambda(root_predicate)

    return pre_filters, post_filter


class Runner:
    """
    Manages processing events and dispatching commands on behalf of a program.
    """

    def __init__(self, loop, client, program, defined_aliases):
        self.pre_filters, self.post_filter = compute_filters(program.select.predicate)

        self._state = _ProgramState(loop, client, program, self.post_filter, defined_aliases)
        self._work_queue = asyncio.Queue(loop=loop)
        self._current_tx = None
        self.future = loop.create_future()

    def state_updater(self, aliases) -> Callable[[Any, Any], None]:
        """
        Return a function that is suitable as a callback to the underlying event stream. On each
        invocation, the (cid, cdata) tuple associated with the event is applied under the specified
        aliases.
        """
        # TODO: A future iteration of this could then selectively apply state updates that match
        # the filter instead of eagerly adding state for all aliases, always
        return partial(self.on_update, aliases)

    def start_transaction(self, _, tdata: TransactionStartEvent):
        # TODO: Framework support for more easily disconnecting event handlers
        if self.future.done():
            return

        self._current_tx = _TransactionBuilder()
        self._current_tx.bim = tdata.command_id

    def on_update(self, aliases, cid, cdata):
        # TODO: Framework support for more easily disconnecting event handlers
        if self.future.done():
            return

        # knock out any data that is not relevant for the resultant cross-product
        for alias in aliases:
            pre_filter_fn = self.pre_filters.get(alias)
            if pre_filter_fn is not None and cdata is not None:
                if not pre_filter_fn({alias: (cid, cdata)}):
                    return

        if self._current_tx is not None:
            self._current_tx.events.extend((alias, cid, cdata) for alias in aliases)
        else:
            LOG.warning('Received an event %s %s outside the context of a transaction!', cid, cdata)

    def end_transaction(self, _, __):
        # TODO: Framework support for more easily disconnecting event handlers
        if self.future.done():
            return

        if self._current_tx is None:
            raise Exception(f'Program {self._state._program.name} callbacks were called out of order!')

        tx = self._current_tx.build()
        self._current_tx = None
        if not self._state.apply_transaction(tx):
            if isinstance(self._state._error, Exception):
                self.future.set_exception(self._state._error)
            elif self._state._error is not None:
                self.future.set_exception(Exception(self._state._error))
            else:
                self.future.set_result(None)

        # self._work_queue.put_nowait(functools.partial(self._state.apply_transaction, tx))

    # def end_block(self, _, __):
    #     # TODO: Framework support for more easily disconnecting event handlers
    #     if self.future.done():
    #         return
    #
    #     # TODO: Why is this needed? _is_ this needed?
    #     self._state._attempt_submit()

    def log_post_checks(self):
        """
        Log warnings based on operations that have (or more helpfully, have NOT) been performed.
        """
        self._state.log_post_checks()

    def run(self) -> Future:
        """
        Return a ``Future`` that completes when the runner has fulfilled all of its work. This
        ``Future`` only completes naturally for `init()`-style programs.
        """
        return self.future


def compute_update(state, predicate_fn, update_fn):
    """
    Creates a list of Ledger API update commands on the given state, filtered by the predicate.

    :param state:
        A dictionary whose keys are aliases and values are dict of cid -> cdata.
    :param predicate_fn:
        A function that, given a state row, returns `True` if the state row is to be considered.
    :param update_fn:
        A function that, given a state row, returns a list of updates.
    """
    pred = functools.partial(filter, predicate_fn)
    cmder = functools.partial(map, update_fn)

    follow_ups = []
    for commands in cmder(pred(projection_iter(state))):
        follow_ups.extend(commands)
    return follow_ups


def projection_iter(state):
    """
    Project a state dict (aliases -> (cid -> cdata)) to the full cross product of all matching
    sub-records.

    :return:
        An iterator over `{alias:row}` for each possible result set derivable from the given state.
    """
    result_set = itertools.product(*[x.items() for x in state.values()])
    for result in result_set:
        yield dict(zip(state.keys(), result))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains pretty-print/formatting utilities.
"""

from dataclasses import dataclass
from typing import Optional


@dataclass(frozen=True)
class PrettyOptions:
    """
    Display options for pretty-printing DAML ASTs.

    Instance attributes:

    .. attribute:: PrettyOptions.column_width

        The maximum number of columns to use when rendering text, or ``None`` if lines should not
        wrap.

    .. attribute:: PrettyOptions.show_hidden_types

        ``True`` to render built-in DAML types defined in ``DA.Internal`` or ``GHC`` and specially
        generated names.

    .. attribute:: PrettyOptions.format

        A string that identifies the target language to render.
    """

    column_width: Optional[int] = None
    show_hidden_types: bool = False
    format: str = 'daml'



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from argparse import ArgumentParser
from .. import __version__ as version
from ._base import CliCommand


class VersionCommand(CliCommand):
    name = 'version'

    def parser(self) -> ArgumentParser:
        return ArgumentParser()

    def execute(self, args) -> int:
        print(version)
        return 0



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

# NOTE TO IMPLEMENTORS: A future version of this file is intended to be code-generated instead of
# manually maintained. The makeup of this file is intentionally highly formulaic in order to
# facilitate a smooth transition to automatically-generated data structures.

from dataclasses import dataclass
from enum import Enum
from io import StringIO
from typing import Callable, Optional, Sequence

from ._base import MISSING, T
from ..model.types import ModuleRef, ValueReference, TypeReference as TypeConName, TypeReference
from ..util.typing import safe_cast


class Unit:
    __slots__ = ()


UNIT = Unit()


# PackageRef


class DottedName:
    __slots__ = 'segments',

    segments: Sequence[str]

    def __init__(self, segments: Sequence[str] = ()):
        object.__setattr__(self, 'segments', tuple(segments))


# ModuleRef
# TypeConName


ValName = ValueReference


@dataclass(frozen=True)
class FieldWithType:
    field: str
    type: 'Type'


@dataclass(frozen=True)
class VarWithType:
    var: str
    type: 'Type'


@dataclass(frozen=True)
class TypeVarWithKind:
    var: str
    kind: 'Kind'

    def __str__(self):
        return f'{self.var} : {self.kind}'

    def __repr__(self):
        return f'TypeVarWithKind({self})'


@dataclass(frozen=True)
class FieldWithExpr:
    field: str
    expr: 'Expr'

    def __repr__(self):
        return f'FieldWithExpr({self.field}={self.expr})'


@dataclass(frozen=True)
class Binding:
    binder: 'VarWithType'
    bound: 'Expr'


class Kind:
    __slots__ = ('_Sum_name', '_Sum_value')

    class Arrow:
        params: 'Sequence[Kind]'
        result: 'Kind'

        def __init__(self, params: 'Sequence[Kind]', result: 'Kind'):
            self.params = params
            self.result = result

    def __init__(
            self,
            star: 'Unit' = MISSING,
            arrow: 'Arrow' = MISSING):
        if star is not MISSING:
            object.__setattr__(self, '_Sum_name', 'star')
            object.__setattr__(self, '_Sum_value', star)
        elif arrow is not MISSING:
            object.__setattr__(self, '_Sum_name', 'arrow')
            object.__setattr__(self, '_Sum_value', arrow)
        else:
            raise ValueError('at least one must be specified')

    @property
    def star(self) -> 'Optional[Unit]':
        return self._Sum_value if self._Sum_name == 'star' else None

    @property
    def arrow(self) -> 'Optional[Arrow]':
        return self._Sum_value if self._Sum_name == 'arrow' else None

    def __repr__(self):
        if self._Sum_name == 'star':
            return '*'
        else:
            arrow = self.arrow
            params = [repr(i) for i in (*arrow.params, arrow.result)]
            return ' -> '.join((f'({i})' if ' ' in i else i for i in params))


class PrimType(Enum):
    MAP_GENERIC = -1  # arity = 2 (this does NOT correspond to a DAML Primitive, but is treated as a
                      # primitive in dazl to allow specialized processing of it
    UNIT = 0          # arity = 0
    BOOL = 1          # arity = 0
    INT64 = 2         # arity = 0
    DECIMAL = 3       # arity = 0
    CHAR = 4          # arity = 0, we have removed this in favor of TEXT for everything text related
    TEXT = 5          # arity = 0
    TIMESTAMP = 6     # arity = 0
    RELTIME = 7       # we removed this in favor of INT64.
    PARTY = 8         # arity = 0
    LIST = 9          # arity = 1
    UPDATE = 10       # arity = 1
    SCENARIO = 11     # arity = 1
    DATE = 12         # arity = 0, days since the unix epoch
    CONTRACT_ID = 13  # arity = 1
    OPTIONAL = 14     # arity = 1
    ARROW = 15        # arity = 2
    MAP = 16          # arity = 1


# noinspection PyShadowingBuiltins
class Type:

    @dataclass(frozen=True)
    class Var:
        var: 'str'
        args: 'Sequence[Type]'

    class Con:
        tycon: 'TypeReference'
        args: 'Sequence[Type]'

        def __init__(self, tycon: 'TypeReference', args: 'Sequence[Type]'):
            self.tycon = tycon
            self.args = tuple(args)

        def __repr__(self):
            return f'Type.Con(tycon={self.tycon}, args={self.args})'

    class Prim:
        prim: 'PrimType'
        args: 'Sequence[Type]'
        internal_type: 'Optional[Type]'

        def __init__(
                self,
                prim: 'PrimType',
                args: 'Sequence[Type]',
                internal_type: 'Optional[Type]' = None):
            self.prim = prim
            self.args = tuple(args)
            self.internal_type = internal_type

    class Forall:
        vars: 'Sequence[TypeVarWithKind]'
        body: 'Type'

        # noinspection PyShadowingBuiltins
        def __init__(self, vars: 'Sequence[TypeVarWithKind]', body: 'Type'):
            self.vars = vars
            self.body = body

    class Tuple:
        fields: 'Sequence[FieldWithType]'

        def __init__(self, fields: 'Sequence[FieldWithType]'):
            self.fields = tuple(fields)

    @dataclass(frozen=True)
    class App:
        tyfun: 'Type'
        args: 'Sequence[Type]'

    __slots__ = '_Sum_name', '_Sum_value'

    def __init__(
            self,
            var: 'Type.Var' = MISSING,
            con: 'Type.Con' = MISSING,
            prim: 'Type.Prim' = MISSING,
            forall: 'Type.Forall' = MISSING,
            tuple: 'Type.Tuple' = MISSING):
        if var is not MISSING:
            object.__setattr__(self, '_Sum_name', 'var')
            object.__setattr__(self, '_Sum_value', var)
        elif con is not MISSING:
            object.__setattr__(self, '_Sum_name', 'con')
            object.__setattr__(self, '_Sum_value', con)
        elif prim is not MISSING:
            object.__setattr__(self, '_Sum_name', 'prim')
            object.__setattr__(self, '_Sum_value', prim)
        elif forall is not MISSING:
            object.__setattr__(self, '_Sum_name', 'forall')
            object.__setattr__(self, '_Sum_value', forall)
        elif tuple is not MISSING:
            object.__setattr__(self, '_Sum_name', 'tuple')
            object.__setattr__(self, '_Sum_value', tuple)
        else:
            raise ValueError('unknown sum type')

    @property
    def var(self) -> 'Type.Var':
        return self._Sum_value if self._Sum_name == 'var' else None

    @property
    def con(self) -> 'Type.Con':
        return self._Sum_value if self._Sum_name == 'con' else None

    @property
    def prim(self) -> 'Type.Prim':
        return self._Sum_value if self._Sum_name == 'prim' else None

    @property
    def forall(self) -> 'Type.Forall':
        return self._Sum_value if self._Sum_name == 'forall' else None

    @property
    def tuple(self) -> 'Type.Tuple':
        return self._Sum_value if self._Sum_name == 'tuple' else None

    # noinspection PyPep8Naming
    def Sum_match(
            self,
            var: 'Callable[[Type.Var], T]',
            con: 'Callable[[Type.Con], T]',
            prim: 'Callable[[Type.Prim], T]',
            forall: 'Callable[[Type.Forall], T]',
            tuple: 'Callable[[Type.Tuple], T]') -> 'T':
        if self._Sum_name == 'var':
            return var(self._Sum_value)
        elif self._Sum_name == 'con':
            return con(self._Sum_value)
        elif self._Sum_name == 'prim':
            return prim(self._Sum_value)
        elif self._Sum_name == 'forall':
            return forall(self._Sum_value)
        elif self._Sum_name == 'tuple':
            return tuple(self._Sum_value)
        else:
            raise Exception('invalid _Sum_name value')

    def __setattr__(self, key, value):
        raise Exception('Type is a read-only object')

    def __str__(self):
        """
        Return a str representation of this :class:`Type`. This is intentionally made to look more
        similar to DAML syntax.
        """
        # This import is done inside the function instead of outside to avoid circular references
        from ..pretty import DAML_PRETTY_PRINTER
        return DAML_PRETTY_PRINTER.visit_type(self)

    def __repr__(self):
        """
        Return a str representation of this :class:`Type`. This is intentionally made to look more
        similar to DAML syntax.
        """
        # This import is done inside the function instead of outside to avoid circular references
        from ..pretty import DAML_PRETTY_PRINTER

        # If rendering the type fails, print _something_; crashing in a repr(...) could be quite
        # obnoxious for users of this library
        # noinspection PyBroadException
        try:
            return f'Type({DAML_PRETTY_PRINTER.visit_type(self)})'
        except:  # noqa
            return 'Type(...)'


class PrimLit:

    __slots__ = ('_Sum_name', '_Sum_value')

    def __init__(
            self,
            int64: int = MISSING,
            decimal: str = MISSING,
            text: str = MISSING,
            timestamp: float = MISSING,
            party: str = MISSING,
            date: int = MISSING):
        if int64 is not MISSING:
            object.__setattr__(self, '_Sum_name', 'int64')
            object.__setattr__(self, '_Sum_value', int64)
        elif decimal is not MISSING:
            object.__setattr__(self, '_Sum_name', 'decimal')
            object.__setattr__(self, '_Sum_value', decimal)
        elif text is not MISSING:
            object.__setattr__(self, '_Sum_name', 'text')
            object.__setattr__(self, '_Sum_value', text)
        elif timestamp is not MISSING:
            object.__setattr__(self, '_Sum_name', 'timestamp')
            object.__setattr__(self, '_Sum_value', timestamp)
        elif party is not MISSING:
            object.__setattr__(self, '_Sum_name', 'party')
            object.__setattr__(self, '_Sum_value', party)
        elif date is not MISSING:
            object.__setattr__(self, '_Sum_name', 'date')
            object.__setattr__(self, '_Sum_value', date)

    @property
    def int64(self) -> 'Optional[int]':
        return self._Sum_value if self._Sum_name == 'int64' else None

    @property
    def decimal(self) -> 'Optional[str]':
        """
        our decimal type would fit in an int128, but sadly Protobuf does not
        have one. so, string it is. note that we can't store the whole and
        decimal part in two numbers either, because 10^28 > 2^63.
        """
        return self._Sum_value if self._Sum_name == 'decimal' else None

    @property
    def text(self) -> 'Optional[str]':
        return self._Sum_value if self._Sum_name == 'text' else None

    @property
    def timestamp(self) -> 'Optional[float]':
        """
        microseconds since the UNIX epoch. can go backwards. fixed
        since the vast majority of values will be greater than
        2^28, since currently the number of microseconds since the
        epoch is greater than that. Range: 0001-01-01T00:00:00Z to
        9999-12-31T23:59:59.999999Z, so that we can convert to/from
        https://www.ietf.org/rfc/rfc3339.txt
        """
        return self._Sum_value if self._Sum_name == 'timestamp' else None

    @property
    def party(self) -> 'Optional[str]':
        return self._Sum_value if self._Sum_name == 'party' else None

    @property
    def date(self) -> 'Optional[int]':
        """
        days since the unix epoch. can go backwards. limited from
        0001-01-01 to 9999-12-31, also to be compatible with
        https://www.ietf.org/rfc/rfc3339.txt
        """
        return self._Sum_value if self._Sum_name == 'date' else None


# noinspection PyShadowingBuiltins
class Location:

    class Range:
        __slots__ = ('start_line', 'start_col', 'end_line', 'end_col')

        def __init__(
                self,
                start_line: int = 0,
                start_col: int = 0,
                end_line: int = 0,
                end_col: int = 0):
            object.__setattr__(self, 'start_line', start_line)
            object.__setattr__(self, 'start_col', start_col)
            object.__setattr__(self, 'end_line', end_line)
            object.__setattr__(self, 'end_col', end_col)

        start_line: int
        start_col: int
        end_line: int
        end_col: int

    module: 'ModuleRef'
    range: 'Range'

    def __init__(self, module: 'ModuleRef' = MISSING, range: 'Range' = MISSING):
        object.__setattr__(self, 'module', module)
        object.__setattr__(self, 'range', range)


# noinspection PyShadowingBuiltins
class Expr:

    class RecCon:
        tycon: 'Type.Con'
        fields: 'Sequence[FieldWithExpr]'

        def __init__(self, tycon: 'Type.Con', fields: 'Sequence[FieldWithExpr]'):
            self.tycon = tycon
            self.fields = tuple(fields)

        def __repr__(self):
            return f'Expr.RecCon(tycon={self.tycon}, fields={self.fields})'

    @dataclass(frozen=True)
    class RecProj:
        tycon: 'Type.Con'  # Always fully applied string
        field: str
        record: 'Expr'

        def __repr__(self):
            with StringIO() as buf:
                buf.write('Expr.RecProj(')
                if self.tycon is not None:
                    buf.write('tycon=')
                    buf.write(repr(self.tycon))
                    buf.write(', ')
                buf.write('record=')
                buf.write(repr(self.record))
                buf.write(', field=')
                buf.write(repr(self.field))
                buf.write(')')
                return buf.getvalue()

    # Set `field` in `record` to `update`.
    @dataclass(frozen=True)
    class RecUpd:
        tycon: 'Type.Con'
        field: str
        record: 'Expr'
        update: 'Expr'

    class VariantCon:
        tycon: 'Type.Con'  # Always fully applied
        variant_con: str
        variant_arg: 'Expr'

        def __init__(self, tycon: 'Type.Con', variant_con: str, variant_arg: 'Expr'):
            self.tycon = tycon
            self.variant_con = variant_con
            self.variant_arg = variant_arg

    @dataclass(frozen=True)
    class EnumCon:
        tycon: 'Type.Con'  # Always fully applied
        enum_con: str

    @dataclass(frozen=True)
    class TupleCon:
        fields: 'Sequence[FieldWithExpr]'  # length > 0

    @dataclass(frozen=True)
    class TupleProj:
        field: str
        tuple: 'Expr'

    # Set `field` in `tuple` to `update`.
    @dataclass(frozen=True)
    class TupleUpd:
        field: str
        tuple: 'Expr'
        update: 'Expr'

    @dataclass(frozen=True)
    class App:
        fun: 'Expr'
        args: 'Sequence[Expr]'  # length > 0

    class TyApp:
        expr: 'Expr'
        types: 'Sequence[Type]'  # length > 0

        def __init__(self, expr: 'Expr', types: 'Sequence[Type]'):
            object.__setattr__(self, 'expr', safe_cast(Expr, expr))
            object.__setattr__(self, 'types', tuple(types))

        def __repr__(self):
            return f'Expr.TyApp(types={self.types}, expr={self.expr})'

    class Abs:
        param: 'Sequence[VarWithType]'  # length > 0
        body: 'Expr'

        def __init__(self, param: 'Sequence[VarWithType]', body: 'Expr'):
            object.__setattr__(self, 'param', param)
            object.__setattr__(self, 'body', safe_cast(Expr, body))

        def __repr__(self):
            with StringIO() as buf:
                buf.write('Expr.Abs(𝜆 ')
                buf.write(' '.join(vwt.var for vwt in self.param))
                buf.write(' . ')
                buf.write(repr(self.body))
                buf.write(')')
                return buf.getvalue()

    @dataclass(frozen=True)
    class TyAbs:
        param: 'Sequence[TypeVarWithKind]'
        body: 'Expr'

    @dataclass(frozen=True)
    class Nil:
        type: 'Type'

        __slots__ = 'type',

        # noinspection PyShadowingBuiltins
        def __init__(self, type: 'Type'):
            object.__setattr__(self, 'type', type)

        def __repr__(self):
            return 'Nil()'

    @dataclass(frozen=True)
    class Cons:
        type: 'Type'
        front: 'Sequence[Expr]'  # length > 0
        tail: 'Expr'

        def __repr__(self):
            with StringIO() as buf:
                buf.write('Expr.Cons(front=')
                buf.write(repr(list(self.front)))
                buf.write(', tail=')
                buf.write(repr(self.tail))
                buf.write(')')
                return buf.getvalue()

    # noinspection PyPep8Naming
    class None_:
        type: 'Type'

        # noinspection PyShadowingBuiltins
        def __init__(self, type: 'Type'):
            self.type = type

        def __repr__(self):
            return f'Expr.None(: {self.type})'

    class Some:
        type: 'Type'
        body: 'Expr'

        # noinspection PyShadowingBuiltins
        def __init__(self, type: 'Type', body: 'Expr'):
            self.type = type
            self.body = body

        def __repr__(self):
            return f'Expr.Some({self.body!r} : {self.type})'

    __slots__ = 'location', '_Sum_name', '_Sum_value'

    def __init__(
            self,
            var: 'str' = MISSING,
            val: 'ValName' = MISSING,
            builtin: 'BuiltinFunction' = MISSING,
            prim_con: 'PrimCon' = MISSING,
            prim_lit: 'PrimLit' = MISSING,
            rec_con: 'RecCon' = MISSING,
            rec_proj: 'RecProj' = MISSING,
            variant_con: 'VariantCon' = MISSING,
            enum_con: 'EnumCon' = MISSING,
            tuple_con: 'TupleCon' = MISSING,
            tuple_proj: 'TupleProj' = MISSING,
            app: 'App' = MISSING,
            ty_app: 'TyApp' = MISSING,
            abs: 'Abs' = MISSING,
            ty_abs: 'TyAbs' = MISSING,
            case: 'Case' = MISSING,
            let: 'Block' = MISSING,
            nil: 'Nil' = MISSING,
            cons: 'Cons' = MISSING,
            update: 'Update' = MISSING,
            scenario: 'Scenario' = MISSING,
            rec_upd: 'RecUpd' = MISSING,
            tuple_upd: 'TupleUpd' = MISSING,
            none: 'None_' = MISSING,
            some: 'Some' = MISSING,
            location: 'Location' = MISSING):
        object.__setattr__(self, 'location', location)
        if var is not MISSING:
            object.__setattr__(self, '_Sum_name', 'var')
            object.__setattr__(self, '_Sum_value', var)
        elif val is not MISSING:
            object.__setattr__(self, '_Sum_name', 'val')
            object.__setattr__(self, '_Sum_value', val)
        elif builtin is not MISSING:
            object.__setattr__(self, '_Sum_name', 'builtin')
            object.__setattr__(self, '_Sum_value', builtin)
        elif prim_con is not MISSING:
            object.__setattr__(self, '_Sum_name', 'prim_con')
            object.__setattr__(self, '_Sum_value', prim_con)
        elif prim_lit is not MISSING:
            object.__setattr__(self, '_Sum_name', 'prim_lit')
            object.__setattr__(self, '_Sum_value', prim_lit)
        elif rec_con is not MISSING:
            object.__setattr__(self, '_Sum_name', 'rec_con')
            object.__setattr__(self, '_Sum_value', rec_con)
        elif rec_proj is not MISSING:
            object.__setattr__(self, '_Sum_name', 'rec_proj')
            object.__setattr__(self, '_Sum_value', rec_proj)
        elif variant_con is not MISSING:
            object.__setattr__(self, '_Sum_name', 'variant_con')
            object.__setattr__(self, '_Sum_value', variant_con)
        elif enum_con is not MISSING:
            object.__setattr__(self, '_Sum_name', 'enum_con')
            object.__setattr__(self, '_Sum_value', enum_con)
        elif tuple_con is not MISSING:
            object.__setattr__(self, '_Sum_name', 'tuple_con')
            object.__setattr__(self, '_Sum_value', tuple_con)
        elif tuple_proj is not MISSING:
            object.__setattr__(self, '_Sum_name', 'tuple_proj')
            object.__setattr__(self, '_Sum_value', tuple_proj)
        elif app is not MISSING:
            object.__setattr__(self, '_Sum_name', 'app')
            object.__setattr__(self, '_Sum_value', app)
        elif ty_app is not MISSING:
            object.__setattr__(self, '_Sum_name', 'ty_app')
            object.__setattr__(self, '_Sum_value', ty_app)
        elif abs is not MISSING:
            object.__setattr__(self, '_Sum_name', 'abs')
            object.__setattr__(self, '_Sum_value', abs)
        elif ty_abs is not MISSING:
            object.__setattr__(self, '_Sum_name', 'ty_abs')
            object.__setattr__(self, '_Sum_value', ty_abs)
        elif case is not MISSING:
            object.__setattr__(self, '_Sum_name', 'case')
            object.__setattr__(self, '_Sum_value', case)
        elif let is not MISSING:
            object.__setattr__(self, '_Sum_name', 'let')
            object.__setattr__(self, '_Sum_value', let)
        elif nil is not MISSING:
            object.__setattr__(self, '_Sum_name', 'nil')
            object.__setattr__(self, '_Sum_value', nil)
        elif cons is not MISSING:
            object.__setattr__(self, '_Sum_name', 'cons')
            object.__setattr__(self, '_Sum_value', cons)
        elif update is not MISSING:
            object.__setattr__(self, '_Sum_name', 'update')
            object.__setattr__(self, '_Sum_value', update)
        elif scenario is not MISSING:
            object.__setattr__(self, '_Sum_name', 'scenario')
            object.__setattr__(self, '_Sum_value', scenario)
        elif rec_upd is not MISSING:
            object.__setattr__(self, '_Sum_name', 'rec_upd')
            object.__setattr__(self, '_Sum_value', rec_upd)
        elif tuple_upd is not MISSING:
            object.__setattr__(self, '_Sum_name', 'tuple_upd')
            object.__setattr__(self, '_Sum_value', tuple_upd)
        elif none is not MISSING:
            object.__setattr__(self, '_Sum_name', 'none')
            object.__setattr__(self, '_Sum_value', none)
        elif some is not MISSING:
            object.__setattr__(self, '_Sum_name', 'some')
            object.__setattr__(self, '_Sum_value', some)
        else:
            raise ValueError(f'At least one valid Sum value must be supplied!')

    def __setattr__(self, key, value):
        raise Exception('Expr is read-only')

    @property
    def var(self) -> 'Optional[str]':
        return self._Sum_value if self._Sum_name == 'var' else None

    @property
    def val(self) -> 'Optional[ValName]':
        return self._Sum_value if self._Sum_name == 'val' else None

    @property
    def builtin(self) -> 'Optional[BuiltinFunction]':
        return self._Sum_value if self._Sum_name == 'builtin' else None

    @property
    def prim_con(self) -> 'Optional[PrimCon]':
        return self._Sum_value if self._Sum_name == 'prim_con' else None

    @property
    def prim_lit(self) -> 'Optional[PrimLit]':
        return self._Sum_value if self._Sum_name == 'prim_lit' else None

    @property
    def rec_con(self) -> 'Optional[Expr.RecCon]':
        return self._Sum_value if self._Sum_name == 'rec_con' else None

    @property
    def rec_proj(self) -> 'Optional[RecProj]':
        return self._Sum_value if self._Sum_name == 'rec_proj' else None

    @property
    def variant_con(self) -> 'Optional[VariantCon]':
        return self._Sum_value if self._Sum_name == 'variant_con' else None

    @property
    def enum_con(self) -> 'Optional[EnumCon]':
        return self._Sum_value if self._Sum_name == 'enum_con' else None

    @property
    def tuple_con(self) -> 'Optional[TupleCon]':
        return self._Sum_value if self._Sum_name == 'tuple_con' else None

    @property
    def tuple_proj(self) -> 'Optional[TupleProj]':
        return self._Sum_value if self._Sum_name == 'tuple_proj' else None

    @property
    def app(self) -> 'Optional[App]':
        return self._Sum_value if self._Sum_name == 'app' else None

    @property
    def ty_app(self) -> 'Optional[Expr.TyApp]':
        return self._Sum_value if self._Sum_name == 'ty_app' else None

    @property
    def abs(self) -> 'Optional[Expr.Abs]':
        return self._Sum_value if self._Sum_name == 'abs' else None

    @property
    def ty_abs(self) -> 'Optional[Expr.TyAbs]':
        return self._Sum_value if self._Sum_name == 'ty_abs' else None

    @property
    def case(self) -> 'Optional[Case]':
        return self._Sum_value if self._Sum_name == 'case' else None

    @property
    def let(self) -> 'Optional[Block]':
        return self._Sum_value if self._Sum_name == 'let' else None

    @property
    def nil(self) -> 'Optional[Expr.Nil]':
        return self._Sum_value if self._Sum_name == 'nil' else None

    @property
    def cons(self) -> 'Optional[Expr.Cons]':
        return self._Sum_value if self._Sum_name == 'cons' else None

    @property
    def update(self) -> 'Optional[Update]':
        return self._Sum_value if self._Sum_name == 'update' else None

    @property
    def scenario(self) -> 'Optional[Scenario]':
        return self._Sum_value if self._Sum_name == 'scenario' else None

    @property
    def rec_upd(self) -> 'Optional[RecUpd]':
        return self._Sum_value if self._Sum_name == 'rec_upd' else None

    @property
    def tuple_upd(self) -> 'Optional[TupleUpd]':
        return self._Sum_value if self._Sum_name == 'tuple_upd' else None

    @property
    def none(self) -> 'Optional[None_]':
        return self._Sum_value if self._Sum_name == 'none' else None

    @property
    def some(self) -> 'Optional[Some]':
        return self._Sum_value if self._Sum_name == 'some' else None

    # noinspection PyPep8Naming
    def Sum_match(
            self,
            var: 'Callable[[str], T]',
            val: 'Callable[[ValName], T]',
            builtin: 'Callable[[BuiltinFunction], T]',
            prim_con: 'Callable[[PrimCon], T]',
            prim_lit: 'Callable[[PrimLit], T]',
            rec_con: 'Callable[[RecCon], T]',
            rec_proj: 'Callable[[RecProj], T]',
            variant_con: 'Callable[[VariantCon], T]',
            enum_con: 'Callable[[EnumCon], T]',
            tuple_con: 'Callable[[TupleCon], T]',
            tuple_proj: 'Callable[[TupleProj], T]',
            app: 'Callable[[App], T]',
            ty_app: 'Callable[[TyApp], T]',
            abs: 'Callable[[Abs], T]',
            ty_abs: 'Callable[[TyAbs], T]',
            case: 'Callable[[Case], T]',
            let: 'Callable[[Block], T]',
            nil: 'Callable[[Nil], T]',
            cons: 'Callable[[Cons], T]',
            update: 'Callable[[Update], T]',
            scenario: 'Callable[[Scenario], T]',
            rec_upd: 'Callable[[RecUpd], T]',
            tuple_upd: 'Callable[[TupleUpd], T]',
            none: 'Callable[[None_], T]',
            some: 'Callable[[Some], T]') -> 'T':
        if self._Sum_name == 'var':
            return var(self.var)
        elif self._Sum_name == 'val':
            return val(self.val)
        elif self._Sum_name == 'builtin':
            return builtin(self.builtin)
        elif self._Sum_name == 'prim_con':
            return prim_con(self.prim_con)
        elif self._Sum_name == 'prim_lit':
            return prim_lit(self.prim_lit)
        elif self._Sum_name == 'rec_con':
            return rec_con(self.rec_con)
        elif self._Sum_name == 'rec_proj':
            return rec_proj(self.rec_proj)
        elif self._Sum_name == 'variant_con':
            return variant_con(self.variant_con)
        elif self._Sum_name == 'enum_con':
            return enum_con(self.enum_con)
        elif self._Sum_name == 'tuple_con':
            return tuple_con(self.tuple_con)
        elif self._Sum_name == 'tuple_proj':
            return tuple_proj(self.tuple_proj)
        elif self._Sum_name == 'app':
            return app(self.app)
        elif self._Sum_name == 'ty_app':
            return ty_app(self.ty_app)
        elif self._Sum_name == 'abs':
            return abs(self.abs)
        elif self._Sum_name == 'ty_abs':
            return ty_abs(self.ty_abs)
        elif self._Sum_name == 'case':
            return case(self.case)
        elif self._Sum_name == 'let':
            return let(self.let)
        elif self._Sum_name == 'nil':
            return nil(self.nil)
        elif self._Sum_name == 'cons':
            return cons(self.cons)
        elif self._Sum_name == 'update':
            return update(self.update)
        elif self._Sum_name == 'scenario':
            return scenario(self.scenario)
        elif self._Sum_name == 'rec_upd':
            return rec_upd(self.rec_upd)
        elif self._Sum_name == 'tuple_upd':
            return tuple_upd(self.tuple_upd)
        elif self._Sum_name == 'none':
            return none(self.none)
        elif self._Sum_name == 'some':
            return some(self.some)
        else:
            raise Exception

    def __repr__(self):
        return f'Expr({self._Sum_name}={self._Sum_value!r})'


class CaseAlt:
    class Variant:
        con: 'TypeConName'
        variant: str
        binder: str

        def __init__(self, con, variant, binder):
            self.con = con
            self.variant = variant
            self.binder = binder

    @dataclass(frozen=True)
    class Cons:
        var_head: str
        var_tail: str

    @dataclass(frozen=True)
    class Some:
        var_body: str

    __slots__ = 'body', '_Sum_name', '_Sum_value'
    body: 'Expr'

    def __init__(
            self,
            default: 'Unit' = MISSING,
            variant: 'Variant' = MISSING,
            prim_con: 'PrimCon' = MISSING,
            nil: 'Unit' = MISSING,
            cons: 'Cons' = MISSING,
            none: 'Unit' = MISSING,
            some: 'Some' = MISSING,
            body: 'Expr' = MISSING):
        object.__setattr__(self, 'body', body)
        if default is not MISSING:
            object.__setattr__(self, '_Sum_name', 'default')
            object.__setattr__(self, '_Sum_value', default)
        elif variant is not MISSING:
            object.__setattr__(self, '_Sum_name', 'variant')
            object.__setattr__(self, '_Sum_value', variant)
        elif prim_con is not MISSING:
            object.__setattr__(self, '_Sum_name', 'prim_con')
            object.__setattr__(self, '_Sum_value', prim_con)
        elif nil is not MISSING:
            object.__setattr__(self, '_Sum_name', 'nil')
            object.__setattr__(self, '_Sum_value', nil)
        elif cons is not MISSING:
            object.__setattr__(self, '_Sum_name', 'cons')
            object.__setattr__(self, '_Sum_value', cons)
        elif none is not MISSING:
            object.__setattr__(self, '_Sum_name', 'none')
            object.__setattr__(self, '_Sum_value', none)
        elif some is not MISSING:
            object.__setattr__(self, '_Sum_name', 'some')
            object.__setattr__(self, '_Sum_value', some)

    @property
    def default(self) -> 'Optional[Unit]':
        return self._Sum_value if self._Sum_name == 'default' else None

    @property
    def variant(self) -> 'Optional[Variant]':
        return self._Sum_value if self._Sum_name == 'variant' else None

    @property
    def prim_con(self) -> 'Optional[PrimCon]':
        return self._Sum_value if self._Sum_name == 'prim_con' else None

    @property
    def nil(self) -> 'Optional[Unit]':
        return self._Sum_value if self._Sum_name == 'nil' else None

    @property
    def cons(self) -> 'Optional[Cons]':
        return self._Sum_value if self._Sum_name == 'cons' else None

    @property
    def none(self) -> 'Optional[Unit]':
        return self._Sum_value if self._Sum_name == 'none' else None

    @property
    def some(self) -> 'Optional[Some]':
        return self._Sum_value if self._Sum_name == 'some' else None

    # noinspection PyPep8Naming
    def Sum_match(
            self,
            default: 'Callable[[Unit], T]',
            variant: 'Callable[[Variant], T]',
            prim_con: 'Callable[[PrimCon], T]',
            nil: 'Callable[[Unit], T]',
            cons: 'Callable[[Cons], T]',
            none: 'Callable[[Unit], T]',
            some: 'Callable[[Some], T]'):
        if self._Sum_name == 'default':
            return default(self.default)
        elif self._Sum_name == 'variant':
            return variant(self.variant)
        elif self._Sum_name == 'prim_con':
            return prim_con(self.prim_con)
        elif self._Sum_name == 'nil':
            return nil(self.nil)
        elif self._Sum_name == 'cons':
            return cons(self.cons)
        elif self._Sum_name == 'none':
            return none(self.none)
        elif self._Sum_name == 'some':
            return some(self.some)
        else:
            raise Exception


@dataclass(frozen=True)
class Case:
    # noinspection SpellCheckingInspection
    scrut: 'Expr'
    alts: 'Sequence[CaseAlt]'  # length > 0


@dataclass(frozen=True)
class Block:
    # A block of bindings and an expression.
    # Encodes a sequence of binds in e.g. a let or update block.
    bindings: 'Sequence[Binding]'  # length > 0
    body: 'Expr'


@dataclass(frozen=True)
class Pure:
    type: 'Type'
    expr: 'Expr'


class Update:

    class Create:
        template: 'TypeConName'
        expr: 'Expr'

        def __init__(self, template: 'TypeConName', expr: 'Expr'):
            self.template = template
            self.expr = expr

    class Exercise:
        template: 'TypeConName'
        choice: str
        cid: 'Expr'
        actor: 'Expr'
        arg: 'Expr'

        def __init__(self, template: 'TypeConName', choice: str, cid: 'Expr', actor: 'Expr',
                     arg: 'Expr'):
            self.template = template
            self.choice = choice
            self.cid = cid
            self.actor = actor
            self.arg = arg

    class Fetch:
        template: 'TypeConName'
        cid: 'Expr'

        def __init__(self, template, cid):
            self.template = template
            self.cid = cid

    class EmbedExpr:
        type: 'Type'  # the expression should be of type `Scenario type`
        body: 'Expr'

        # noinspection PyShadowingBuiltins
        def __init__(self, type: 'Type', body: 'Expr'):
            self.type = type
            self.body = body

    class RetrieveByKey:
        template: 'TypeConName'
        key: 'Expr'

        def __init__(self, template: 'TypeConName', key: 'Expr'):
            self.template = template
            self.key = key

    __slots__ = '_Sum_name', '_Sum_value'

    def __init__(
            self,
            pure: 'Pure' = MISSING,
            block: 'Block' = MISSING,
            create: 'Create' = MISSING,
            exercise: 'Exercise' = MISSING,
            fetch: 'Fetch' = MISSING,
            get_time: 'Unit' = MISSING,
            lookup_by_key: 'RetrieveByKey' = MISSING,
            fetch_by_key: 'RetrieveByKey' = MISSING,
            embed_expr: 'EmbedExpr' = MISSING):
        if pure is not MISSING:
            object.__setattr__(self, '_Sum_name', 'pure')
            object.__setattr__(self, '_Sum_value', pure)
        elif block is not MISSING:
            object.__setattr__(self, '_Sum_name', 'block')
            object.__setattr__(self, '_Sum_value', block)
        elif create is not MISSING:
            object.__setattr__(self, '_Sum_name', 'create')
            object.__setattr__(self, '_Sum_value', create)
        elif exercise is not MISSING:
            object.__setattr__(self, '_Sum_name', 'exercise')
            object.__setattr__(self, '_Sum_value', exercise)
        elif fetch is not MISSING:
            object.__setattr__(self, '_Sum_name', 'fetch')
            object.__setattr__(self, '_Sum_value', fetch)
        elif get_time is not MISSING:
            object.__setattr__(self, '_Sum_name', 'get_time')
            object.__setattr__(self, '_Sum_value', get_time)
        elif embed_expr is not MISSING:
            object.__setattr__(self, '_Sum_name', 'embed_expr')
            object.__setattr__(self, '_Sum_value', embed_expr)
        elif lookup_by_key is not MISSING:
            object.__setattr__(self, '_Sum_name', 'lookup_by_key')
            object.__setattr__(self, '_Sum_value', lookup_by_key)
        elif fetch_by_key is not MISSING:
            object.__setattr__(self, '_Sum_name', 'fetch_by_key')
            object.__setattr__(self, '_Sum_value', fetch_by_key)

    @property
    def pure(self) -> 'Optional[Pure]':
        """this is purely for compact serialization -- specifically to
        reduce the AST depth. it adds no expressive power."""
        return self._Sum_value if self._Sum_name == 'pure' else None

    @property
    def block(self) -> 'Optional[Block]':
        return self._Sum_value if self._Sum_name == 'block' else None

    @property
    def create(self) -> 'Optional[Create]':
        return self._Sum_value if self._Sum_name == 'create' else None

    @property
    def exercise(self) -> 'Optional[Exercise]':
        return self._Sum_value if self._Sum_name == 'exercise' else None

    @property
    def fetch(self) -> 'Optional[Fetch]':
        return self._Sum_value if self._Sum_name == 'fetch' else None

    @property
    def get_time(self) -> 'Optional[Unit]':
        return self._Sum_value if self._Sum_name == 'get_time' else None

    @property
    def embed_expr(self) -> 'Optional[EmbedExpr]':
        """see similar constructor in `Scenario` on why this is useful."""
        return self._Sum_value if self._Sum_name == 'embed_expr' else None


class Scenario:
    class Commit:
        party: 'Expr'
        expr: 'Expr'
        ret_type: 'Type'

        def __init__(self, party: 'Expr', expr: 'Expr', ret_type: 'Type'):
            self.party = party
            self.expr = expr
            self.ret_type = ret_type

    class EmbedExpr:
        type: 'Type'  # the expression should be of type `Scenario type`
        body: 'Expr'

        # noinspection PyShadowingBuiltins
        def __init__(self, type: 'Type', body: 'Expr'):
            self.type = type
            self.body = body

    __slots__ = '_Sum_name', '_Sum_value'

    def __init__(
            self,
            pure: 'Pure' = MISSING,
            block: 'Block' = MISSING,
            commit: 'Commit' = MISSING,
            must_fail_at: 'Commit' = MISSING,
            pass_: 'Expr' = MISSING,
            get_time: 'Unit' = MISSING,
            get_party: 'Expr' = MISSING,
            embed_expr: 'EmbedExpr' = MISSING):
        if pure is not MISSING:
            object.__setattr__(self, '_Sum_name', 'pure')
            object.__setattr__(self, '_Sum_value', pure)
        elif block is not MISSING:
            object.__setattr__(self, '_Sum_name', 'block')
            object.__setattr__(self, '_Sum_value', block)
        elif commit is not MISSING:
            object.__setattr__(self, '_Sum_name', 'commit')
            object.__setattr__(self, '_Sum_value', commit)
        elif must_fail_at is not MISSING:
            object.__setattr__(self, '_Sum_name', 'must_fail_at')
            object.__setattr__(self, '_Sum_value', must_fail_at)
        elif pass_ is not MISSING:
            object.__setattr__(self, '_Sum_name', 'pass')
            object.__setattr__(self, '_Sum_value', pass_)
        elif get_time is not MISSING:
            object.__setattr__(self, '_Sum_name', 'get_time')
            object.__setattr__(self, '_Sum_value', get_time)
        elif get_party is not MISSING:
            object.__setattr__(self, '_Sum_name', 'get_party')
            object.__setattr__(self, '_Sum_value', get_party)
        elif embed_expr is not MISSING:
            object.__setattr__(self, '_Sum_name', 'embed_expr')
            object.__setattr__(self, '_Sum_value', embed_expr)

    @property
    def pure(self) -> 'Optional[Pure]':
        """this is purely for compact serialization -- specifically to
        reduce the AST depth. it adds no expressive power."""
        return self._Sum_value if self._Sum_name == 'pure' else None

    @property
    def block(self) -> 'Optional[Block]':
        return self._Sum_value if self._Sum_name == 'block' else None

    @property
    def commit(self) -> 'Optional[Commit]':
        return self._Sum_value if self._Sum_name == 'create' else None

    @property
    def must_fail_at(self) -> 'Optional[Commit]':
        return self._Sum_value if self._Sum_name == 'mustFailAt' else None

    @property
    def pass_(self) -> 'Optional[Expr]':
        return self._Sum_value if self._Sum_name == 'pass' else None

    @property
    def get_time(self) -> 'Optional[Unit]':
        """The expression is of type `Text`."""
        return self._Sum_value if self._Sum_name == 'get_time' else None

    @property
    def get_party(self) -> 'Optional[Expr]':
        return self._Sum_value if self._Sum_name == 'get_party' else None

    @property
    def embed_expr(self) -> 'Optional[EmbedExpr]':
        """
        Embed an expression of type Scenario. note that this construct is useful
        to explicitly mark the start of scenario execution, which is useful in
        top level definitions. for example if we hav

        def test : Scenario Unit = if <blah> then <this> else <that>

        this is not a value, since it's headed with an `if`, but we can turn
        it into a value by wrapping the `if` with this constructor. in that
        case, the `if` will be executed every time the scenario runs --
        as expected.
        """
        return self._Sum_value if self._Sum_name == 'embed_expr' else None


class BuiltinFunction(Enum):
    CONCAT_LIST = -1000

    ADD_DECIMAL = 0
    SUB_DECIMAL = 1
    MUL_DECIMAL = 2
    DIV_DECIMAL = 3
    ROUND_DECIMAL = 6

    ADD_INT64 = 7
    SUB_INT64 = 8
    MUL_INT64 = 9
    DIV_INT64 = 10
    MOD_INT64 = 11
    EXP_INT64 = 12

    FOLDL = 20
    FOLDR = 21

    MAP_EMPTY = 96
    MAP_INSERT = 97
    MAP_LOOKUP = 98
    MAP_DELETE = 99
    MAP_TO_LIST = 100
    MAP_SIZE = 101

    EXPLODE_TEXT = 23
    APPEND_TEXT = 24

    ERROR = 25

    LEQ_INT64 = 33
    LEQ_DECIMAL = 34
    LEQ_TEXT = 36
    LEQ_TIMESTAMP = 37
    LEQ_DATE = 67
    LEQ_PARTY = 89

    LESS_INT64 = 39
    LESS_DECIMAL = 40
    LESS_TEXT = 42
    LESS_TIMESTAMP = 43
    LESS_DATE = 68
    LESS_PARTY = 90

    GEQ_INT64 = 45
    GEQ_DECIMAL = 46
    GEQ_TEXT = 48
    GEQ_TIMESTAMP = 49
    GEQ_DATE = 69
    GEQ_PARTY = 91

    GREATER_INT64 = 51
    GREATER_DECIMAL = 52
    GREATER_TEXT = 54
    GREATER_TIMESTAMP = 55
    GREATER_DATE = 70
    GREATER_PARTY = 92

    TO_TEXT_INT64 = 57
    TO_TEXT_DECIMAL = 58
    TO_TEXT_TEXT = 60
    TO_TEXT_TIMESTAMP = 61
    TO_TEXT_DATE = 71
    TO_QUOTED_TEXT_PARTY = 63  # legacy, remove in next major version
    TO_TEXT_PARTY = 94  # Available Since version 1.2
    TO_TEXT_CODE_POINTS = 105
    FROM_TEXT_PARTY = 95  # Available Since version 1.2
    FROM_TEXT_INT64 = 103  # Available Since version 1.5
    FROM_TEXT_DECIMAL = 104  # Available Since version 1.5
    FROM_TEXT_CODE_POINTS = 106
    SHA256_TEXT = 93  # Available Since version 1.2

    DATE_TO_UNIX_DAYS = 72  # Date -> Int64
    UNIX_DAYS_TO_DATE = 73  # Int64 -> Date

    TIMESTAMP_TO_UNIX_MICROSECONDS = 74  # Timestamp -> Int64
    UNIX_MICROSECONDS_TO_TIMESTAMP = 75  # Int64 -> Timestamp

    INT64_TO_DECIMAL = 76
    DECIMAL_TO_INT64 = 77

    IMPLODE_TEXT = 78

    EQUAL_INT64 = 79
    EQUAL_DECIMAL = 80
    EQUAL_TEXT = 81
    EQUAL_TIMESTAMP = 82
    EQUAL_DATE = 83
    EQUAL_PARTY = 84
    EQUAL_BOOL = 85
    EQUAL_CONTRACT_ID = 86
    EQUAL_LIST = 87

    TRACE = 88

    COERCE_CONTRACT_ID = 102


class PrimCon(Enum):
    CON_UNIT = 0
    CON_FALSE = 1
    CON_TRUE = 2


@dataclass(frozen=True)
class TemplateChoice:
    """Template choice definition."""

    # Name of the choice.
    name: str

    # Indicator whether exercising the choice consumes the contract instance.
    consuming: bool

    # The controllers of the choice. They have type `List Party` and the
    # template parameter in scope, but not the choice parameter. All of these
    # controllers need to authorize the exercising of this choice (aka
    # conjunctive choice controllers).
    controllers: 'Expr'

    # Name to which the choice argument is bound and its type.
    arg_binder: 'VarWithType'

    # Return type of the choice.
    ret_type: 'Type'

    # Follow-up update of the choice. It has type `Update <ret_type>` and both
    # the template parameter and the choice parameter in scope.
    update: 'Expr'

    # Name to bind the ContractId of the contract this choice is exercised on to.
    self_binder: str

    location: 'Location'


@dataclass(frozen=True)
class DefTemplate:
    """Contract template definition"""

    # The type constructor for the template, acting as both
    # the name of the template and the type of the template argument.
    tycon: 'DottedName'

    # Name to which the template argument is bound.
    param: str

    # Optional pre-condition that the template argument must satisfy.
    # When present, it has type `Bool` and the template parameter in scope.
    precond: 'Expr'

    # The signatories of the contract. They have type `List Party` and the
    # template parameter in scope.
    signatories: 'Expr'

    # The agreement text associated with the contract. It has type `Text` and
    # the template parameter in scope.
    agreement: 'Expr'

    # The choices available in the resulting contract.
    choices: 'Sequence[TemplateChoice]'

    # The observers of the contract. They have type `List Party` and the
    # template parameter in scope.
    observers: 'Expr'

    location: 'Location'


class DefDataType:

    class Fields:
        fields: 'Sequence[FieldWithType]'

        def __init__(self, fields: 'Sequence[FieldWithType]'):
            self.fields = fields

    name: DottedName
    params: 'Sequence[TypeVarWithKind]'
    _DataCons_name: str
    _DataCons_value: 'Fields'
    serializable: bool
    location: 'Location'

    def __init__(
            self,
            name: DottedName = MISSING,
            params: 'Sequence[TypeVarWithKind]' = MISSING,
            record: 'DefDataType.Fields' = MISSING,
            variant: 'DefDataType.Fields' = MISSING,
            serializable: bool = MISSING,
            location: 'Location' = MISSING):
        self.name = name
        self.params = params
        if record is not MISSING:
            self._DataCons_name = 'record'
            self._DataCons_value = record
        elif variant is not MISSING:
            self._DataCons_name = 'variant'
            self._DataCons_value = variant
        self.serializable = serializable
        self.location = location

    @property
    def record(self):
        return self._DataCons_value if self._DataCons_name == 'record' else None

    @property
    def variant(self):
        return self._DataCons_value if self._DataCons_name == 'variant' else None


@dataclass(frozen=True)
class DefValue:
    """Value definition"""

    class NameWithType:
        """
        The reason why we have this type instead of just flattening name and type in
        DefValue is that it was VarWithType before, and we want to be binary-compatible
        with it.
        """
        name: 'Sequence[str]'
        type: 'Type'

        # noinspection PyShadowingBuiltins
        def __init__(self, name: 'Sequence[str]', type: 'Type'):
            self.name = tuple(name)
            self.type = type

    name_with_type: 'DefValue.NameWithType'
    expr: 'Expr'

    # If true, the value must not contain any party literals and not reference
    # values which contain party literals.
    # This flag is used to simplify package validation by not requiring an
    # inference but only a check. Such a check must validate that this flag is
    # set correctly and that templates do not reference values which have this
    # flag set to false.
    no_party_literals: bool
    is_test: bool
    location: 'Optional[Location]'


@dataclass(frozen=True)
class FeatureFlags:
    forbidPartyLiterals: bool
    dontDivulgeContractIdsInCreateArguments: bool
    dontDiscloseNonConsumingChoicesToObservers: bool


@dataclass(frozen=True)
class Module:
    name: 'DottedName'
    flags: 'FeatureFlags'
    data_types: 'Sequence[DefDataType]'
    values: 'Sequence[DefValue]'
    templates: 'Sequence[DefTemplate]'


@dataclass(frozen=True)
class Package:
    modules: 'Sequence[Module]'


@dataclass(frozen=True)
class Archive:
    hash: str
    package: 'Package'



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import inspect
from typing import List, Union

from .expr_base import Expr, TRIGGER_ON_INIT, TRIGGER_ANY_TIME
from .expr_impl import TemplateExpr, TemplateSelectsExpr, TemplateSelectExpr, \
    UpdateBlockExpr, UpdateStatementExpr, ProgramExpr


UpdateType = Union[UpdateBlockExpr, UpdateStatementExpr, List[Union[UpdateBlockExpr, UpdateStatementExpr]]]


def from_(*templates: Union[TemplateExpr, TemplateSelectExpr], name=None) -> 'ProgramExprBuilder':
    """
    Builder method for a `ProgramExpr` that runs on a set of templates at any time.
    """
    if name is None:
        frameinfo: inspect.FrameInfo = inspect.stack()[1]
        name = frameinfo.function
    return ProgramExprBuilder(name=name, trigger=TRIGGER_ANY_TIME, templates=templates)


def init(name: str=None) -> 'InitProgramExprBuilder':
    """
    Builder method for a `ProgramExpr` that runs only when an initial condition has been met.
    """
    if name is None:
        frameinfo: inspect.FrameInfo = inspect.stack()[1]
        name = frameinfo.function
    return InitProgramExprBuilder(name)


class InitProgramExprBuilder(object):
    """
    Builder class for a `ProgramExpr`, starting from the assumption that the trigger condition is
    `TRIGGER_ON_INIT`.
    """
    def __init__(self, name: str=None):
        if name is None:
            frameinfo: inspect.FrameInfo = inspect.stack()[1]
            name = frameinfo.function
        self.name = name

    def from_(self, *templates: Union[TemplateExpr, TemplateSelectExpr]) -> 'ProgramExprBuilder':
        """
        Specify that the initial condition is, in part, based on ledger state of a select set of
        templates.
        """
        return ProgramExprBuilder(name=self.name, trigger=TRIGGER_ON_INIT, templates=templates)

    def update(self, expression: Union['UpdateBlockExpr', 'UpdateStatementExpr']) -> 'ProgramExpr':
        """
        Specify the update block to invoke when the ledger is initialized and otherwise has no
        pre-existing state.
        """
        return ProgramExpr(name=self.name, select=None, update=expression, trigger=TRIGGER_ON_INIT)


class ProgramExprBuilder(object):
    def __init__(self, name, trigger, templates, predicate=None):
        self.name = name
        self.trigger = trigger
        self.templates = templates
        self.predicate = predicate

    def where(self, *predicates: 'Expr') -> 'ProgramExprBuilder':
        from . import and_
        return ProgramExprBuilder(
            name=self.name,
            trigger=self.trigger,
            templates=self.templates,
            predicate=and_(*predicates))

    def update(self, expression: UpdateType) -> ProgramExpr:
        """
        Return the :class:`ProgramExpr` with the specified parameters.
        """
        if isinstance(expression, UpdateStatementExpr):
            expression = UpdateBlockExpr((expression,))
        return ProgramExpr(
            name=self.name,
            select=TemplateSelectsExpr(self.templates, self.predicate),
            update=expression,
            trigger=self.trigger)



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Utility methods operating on a full expression tree.
"""

from collections import defaultdict
from typing import Set, Dict

from ._rewriter import RewriteContext
from ..model.types import Type


def get_distinct_templates(expr) -> Dict[str, Set[Type]]:
    """
    Return the distinct :class:`TemplateRef` that occur in the expression tree and their
    associated aliases.

    :return:
        A dict whose keys are template ID strings and values are lists of strings
        that are their corresponding aliases in the expression tree.
    """
    from .expr_impl import TemplateSelectExpr

    template_refs = defaultdict(set)

    context = RewriteContext()

    @context.register(TemplateSelectExpr)
    def _capture_select(expr_):
        template_refs[expr_.template.template_ref.name].add(expr_.alias)
        return expr_

    context.rewrite(expr)

    return {key: sorted(value) for key, value in template_refs.items()}


def get_distinct_aliases(expr) -> Set[str]:
    from .expr_impl import TemplateSelectExpr

    aliases = set()

    context = RewriteContext()

    @context.register(TemplateSelectExpr)
    def _capture_select(expr_: TemplateSelectExpr):
        aliases.add(expr_.alias)
        return expr_

    context.rewrite(expr)

    return aliases



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains :class:`LedgerCapturePlugin` and its supporting infrastructure.
"""



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains utilities for fetching DAML SDK components.
"""

import platform
import tarfile
from dataclasses import dataclass
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory
from typing import Union, Tuple
from urllib.parse import urlparse
from urllib.request import urlretrieve


from .artifacts import load_artifacts, get_artifact_url, parse_component_identifier

from semver import VersionInfo

from .. import LOG


@dataclass(frozen=True)
class FetchResult:
    path: Path
    url: str
    downloaded: bool


def sdk_component_path(component: str) -> Tuple[str, Path]:
    name, version = parse_component_identifier(component, use_default_version=True)
    return name, ensure_sdk_component(name, version).path


def ensure_sdk_component(
        component: str,
        version: Union[None, str, VersionInfo],
        force: bool = False) -> FetchResult:
    """
    Ensures the existence of an SDK component. This may download the component first.
    """
    artifacts = load_artifacts()
    target_platform = 'osx' if platform.system() == 'Darwin' else 'linux'

    artifact_loc = artifacts.artifacts[component]
    if version is not None and isinstance(version, str):
        version = VersionInfo.parse(version)
    elif not isinstance(version, VersionInfo):
        version = artifact_loc.default_version
        if version is None:
            raise ValueError(f'the component {component!r} does not have a default version defined')

    url = get_artifact_url(artifacts.root_url, artifact_loc, version, platform=target_platform)

    installation_dir = Path.home() / '.da' / 'packages' / component / str(version)
    if installation_dir.exists():
        # if the top-level directory already exists, assume we already downloaded it
        if not force:
            return FetchResult(installation_dir, url, False)
    else:
        installation_dir.parent.mkdir(parents=True, exist_ok=True)

    if artifact_loc.packaging in ('tar.gz', 'tgz'):
        extract_sdk_tgz(url, installation_dir, force=force)
    elif artifact_loc.packaging is None or artifact_loc.packaging == 'jar':
        download_package(url, installation_dir)
    else:
        LOG.warning('The packaging type %s does not have a defined mechanism for handling',
                    artifact_loc.packaging)
        download_package(url, installation_dir)

    return FetchResult(installation_dir, url, True)


def extract_sdk_tgz(url: str, target_directory: Path, force: bool = False):
    if not force and target_directory.exists():
        raise RuntimeError(f'the directory already exists: {target_directory}')

    temp_dir = Path.home() / '.da' / 'tmp'
    if not temp_dir.exists():
        temp_dir.mkdir(parents=True, exist_ok=True)

    with TemporaryDirectory(dir=temp_dir) as td:
        tempdir = Path(td)

        download_package(url, tempdir)
        pkg = next(tempdir.iterdir())

        untarred = Path(tempdir) / "extract"
        with tarfile.open(pkg) as tar:
            tar.extractall(untarred)

        rmtree(target_directory, ignore_errors=True)
        next(untarred.iterdir()).rename(target_directory)

        # a little Mojave hack to work around bad packaging of Haskell components
        bad_lib = target_directory / "lib" / "libSystem.B.dylib"
        if bad_lib.exists():
            bad_lib.unlink()


def download_package(url: str, output_directory: Path) -> Path:
    if not output_directory.exists():
        output_directory.mkdir(parents=True)

    p = urlparse(url)
    filename = p.path.rpartition('/')[-1]
    path = output_directory / filename

    try:
        urlretrieve(url, path)
    except:
        LOG.exception('Failed to download %s', url)
        raise

    return path



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import command_service_pb2 as com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class CommandServiceStub(object):
  """Command Service is able to correlate submitted commands with completion data, identify timeouts, and return contextual
  information with each tracking result. This supports the implementation of stateless clients.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.SubmitAndWait = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.CommandService/SubmitAndWait',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
        response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
    self.SubmitAndWaitForTransactionId = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.CommandService/SubmitAndWaitForTransactionId',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionIdResponse.FromString,
        )
    self.SubmitAndWaitForTransaction = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.CommandService/SubmitAndWaitForTransaction',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionResponse.FromString,
        )
    self.SubmitAndWaitForTransactionTree = channel.unary_unary(
        '/com.digitalasset.ledger.api.v1.CommandService/SubmitAndWaitForTransactionTree',
        request_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
        response_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionTreeResponse.FromString,
        )


class CommandServiceServicer(object):
  """Command Service is able to correlate submitted commands with completion data, identify timeouts, and return contextual
  information with each tracking result. This supports the implementation of stateless clients.
  """

  def SubmitAndWait(self, request, context):
    """Submits a single composite command and waits for its result.
    Returns ``RESOURCE_EXHAUSTED`` if the number of in-flight commands reached the maximum (if a limit is configured).
    Propagates the gRPC error of failed submissions including DAML interpretation errors.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubmitAndWaitForTransactionId(self, request, context):
    """Submits a single composite command, waits for its result, and returns the transaction id.
    Returns ``RESOURCE_EXHAUSTED`` if the number of in-flight commands reached the maximum (if a limit is configured).
    Propagates the gRPC error of failed submissions including DAML interpretation errors.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubmitAndWaitForTransaction(self, request, context):
    """Submits a single composite command, waits for its result, and returns the transaction.
    Returns ``RESOURCE_EXHAUSTED`` if the number of in-flight commands reached the maximum (if a limit is configured).
    Propagates the gRPC error of failed submissions including DAML interpretation errors.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SubmitAndWaitForTransactionTree(self, request, context):
    """Submits a single composite command, waits for its result, and returns the transaction tree.
    Returns ``RESOURCE_EXHAUSTED`` if the number of in-flight commands reached the maximum (if a limit is configured).
    Propagates the gRPC error of failed submissions including DAML interpretation errors.
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_CommandServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'SubmitAndWait': grpc.unary_unary_rpc_method_handler(
          servicer.SubmitAndWait,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
          response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
      ),
      'SubmitAndWaitForTransactionId': grpc.unary_unary_rpc_method_handler(
          servicer.SubmitAndWaitForTransactionId,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionIdResponse.SerializeToString,
      ),
      'SubmitAndWaitForTransaction': grpc.unary_unary_rpc_method_handler(
          servicer.SubmitAndWaitForTransaction,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionResponse.SerializeToString,
      ),
      'SubmitAndWaitForTransactionTree': grpc.unary_unary_rpc_method_handler(
          servicer.SubmitAndWaitForTransactionTree,
          request_deserializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
          response_serializer=com_dot_digitalasset_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionTreeResponse.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'com.digitalasset.ledger.api.v1.CommandService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from ..model.core import RunLevel


class RunState:
    """
    Manage the current run level of the client.
    """
    def __init__(self, level=RunLevel.RUN_FOREVER):
        self.level = level

    def handle_sigint(self):
        self.level = RunLevel.TERMINATE_GRACEFULLY

    def handle_sigquit(self):
        self.level = RunLevel.TERMINATE_IMMEDIATELY



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Types that describe the behavior of the ledger itself.
"""

import abc
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any

from .types_store import PackageStore
from .writing import Serializer


class TimeModel(abc.ABC):
    """
    Definition for a time model.
    """

    @abc.abstractmethod
    def get_time(self) -> datetime:
        """
        Retrieve the current time.
        """


class StaticTimeModel(TimeModel):
    """
    Time model backed by a static clock that only advances in time in response to an explicit
    request.
    """

    def __init__(self, start_time: datetime):
        if not isinstance(start_time, datetime):
            raise TypeError('start_time must be a datetime')
        self.current_time = start_time

    def get_time(self) -> datetime:
        return self.current_time


class RealTimeModel(TimeModel):

    def get_time(self) -> datetime:
        """
        Retrieve the current time.
        """
        return datetime.utcnow().replace(tzinfo=timezone.utc)


@dataclass(frozen=True)
class LedgerMetadata:
    """
    Attributes that are invariant with respect to any party on the ledger.
    """
    ledger_id: str
    store: PackageStore
    time_model: TimeModel
    serializer: Serializer[Any, Any]
    protocol_version: str



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Sequence


def unmangle_name(s: str) -> str:
    """
    Convert a name with ``$uXXXX`` character runs into their original forms, which may not
    necessarily be safe DAML identifiers.

    :param s: The string to interpret.
    :return: A new string with the appropriate substitutions.
    """
    return s.replace('$u002b', '+').replace('$u005b', '[').replace('$u005d', ']'). \
        replace('$u003c', '<').replace('$u003e', '>').replace('$u003a', ':'). \
        replace('$u0022', '"').replace('$u0028', '(').replace('$u0029', ')'). \
        replace('$u002f', '/').replace('$u002c', ',').replace('$u003d', '=')


def is_hidden_module_name(name: 'Sequence[str]') -> bool:
    if len(name) == 1 and name[0] in ('GhcPrim', 'Prelude'):
        return True
    if len(name) >= 2:
        if name[0] == 'GHC':
            return True
        elif name[0] == 'DA' and name[1] == 'Internal':
            return True
        elif name[0] == 'Control' and name[1] == 'Exception':
            return True
    return False


def maybe_parentheses(obj, operator: str = ' ') -> str:
    """
    Wrap the string (or object's string representation) in parentheses, but only if required.
    :param obj:
        The object to (potentially) wrap up in parentheses. If it is not a string, it is converted
        to a string first.
    :param operator:
        The operator that needs precedence rules clarified. The default is ``' '`` (a space).
    :return: The (possibly) parenthesized string.
    """
    s = str(obj)
    if operator not in s:
        return s

    # if there are already wrapping punctuation marks, there may not be a need to add additional
    # ones
    open_mark = s[0]
    if open_mark in '([{':
        groupings = [s[0]]
        for c in s[1:-1]:
            if c in '([{':
                groupings.append(c)
            elif c == '}':
                if groupings[-1] == '{':
                    groupings.pop()
                else:
                    groupings.clear()
            elif c == ']':
                if groupings[-1] == '[':
                    groupings.pop()
                else:
                    groupings.clear()
            elif c == ')':
                if groupings[-1] == '(':
                    groupings.pop()
                else:
                    groupings.clear()

            if not groupings:
                # we balanced out all groupings (or just his a syntax error in unmatched groupings);
                # add clarifying parentheses
                return '(' + s + ')'
        if (groupings[0] == '(' and groupings[-1] == ')') or \
           (groupings[0] == '[' and groupings[-1] == ']') or \
           (groupings[0] == '{' and groupings[-1] == '}'):
            return s
    return '(' + s + ')' if operator in s else s


def indent(text: str, spaces: int):
    """
    Prepend every line of the specified text with a set number of spaces. Line endings are
    preserved.
    """
    prefix = ' ' * spaces
    return ''.join(prefix + t for t in text.splitlines(keepends=True))



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Conversion methods to Ledger API Protobuf-generated types from dazl/Pythonic types.
"""
from datetime import date, datetime, timedelta, timezone
from typing import Any, Tuple

from ... import LOG
# noinspection PyPep8Naming
from . import model as G
from ...model.core import ContractId
from ...model.types import VariantType, RecordType, ListType, ContractIdType, \
    UnsupportedType, TemplateChoice, TypeReference, TypeEvaluationContext, SCALAR_TYPE_UNIT, \
    MapType, OptionalType
from ...model.writing import AbstractSerializer, CommandPayload
from ...util.prim_types import to_boolean, to_date, to_datetime, to_decimal, to_int, to_str, \
    decode_variant_dict

# noinspection PyPackageRequirements
from google.protobuf.empty_pb2 import Empty
# noinspection PyPackageRequirements
from google.protobuf.timestamp_pb2 import Timestamp


R = Tuple[str, Any]


_NANOS_PER_MICROSECOND = 1000
_SECONDS_PER_DAY = 24 * 3600


def as_api_timestamp(dt: datetime) -> Timestamp:
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)
    td = dt - datetime(1970, 1, 1, tzinfo=timezone.utc)
    ts = Timestamp()
    ts.seconds = td.seconds + td.days * _SECONDS_PER_DAY
    ts.nanos = td.microseconds * _NANOS_PER_MICROSECOND
    return ts


def pb_get_date(obj: date) -> int:
    """
    Return the Python ``datetime.date`` as the number of days since 1970 Jan 1.
    """
    return (obj - date(1970, 1, 1)).days


def pb_get_timestamp(obj: datetime) -> float:
    """
    Return the Python ``datetime.datetime`` as number of microseconds since
    1970 Jan 1 midnight, UTC.
    """
    if obj.tzinfo is not None and obj.tzinfo.utcoffset(obj) is not None:
        # aware time; translate to UTC
        obj = obj.astimezone(timezone.utc)
    obj = obj.replace(tzinfo=None)
    td: timedelta = obj - datetime(1970, 1, 1, 0, 0, 0)
    return (td.days * 86400 + td.seconds) * 10 ** 6 + td.microseconds


class ProtobufSerializer(AbstractSerializer[G.Command, R]):

    ################################################################################################
    # COMMAND serializers
    ################################################################################################

    def serialize_command_request(self, command_payload: CommandPayload) -> G.SubmitRequest:
        commands = [self.serialize_command(command) for command in command_payload.commands]
        return G.SubmitRequest(commands=G.Commands(
            ledger_id=command_payload.ledger_id,
            workflow_id=command_payload.workflow_id,
            application_id=command_payload.application_id,
            command_id=command_payload.command_id,
            party=command_payload.party,
            commands=commands,
            ledger_effective_time=as_api_timestamp(command_payload.ledger_effective_time),
            maximum_record_time=as_api_timestamp(command_payload.maximum_record_time)))

    def serialize_create_command(self, template_type: RecordType, template_args: R) -> G.Command:
        _, value = template_args

        cmd = G.CreateCommand()
        cmd.template_id.package_id = template_type.name.module.package_id
        cmd.template_id.name = template_type.name.full_name
        cmd.template_id.module_name = '.'.join(template_type.name.module.module_name)
        cmd.template_id.entity_name = '.'.join(template_type.name.name)
        cmd.create_arguments.MergeFrom(value)
        return G.Command(create=cmd)

    def serialize_exercise_command(
            self, contract_id: ContractId, choice_info: TemplateChoice, choice_args: R) \
            -> G.Command:
        type_ref = contract_id.template_id  # type: TypeReference
        ctor, value = choice_args

        cmd = G.ExerciseCommand()
        cmd.template_id.package_id = type_ref.module.package_id
        cmd.template_id.name = type_ref.full_name
        cmd.contract_id = contract_id.contract_id
        cmd.choice = choice_info.name
        if ctor == 'unit':
            cmd.choice_argument.unit.SetInParent()
        else:
            getattr(cmd.choice_argument, ctor).MergeFrom(value)
        return G.Command(exercise=cmd)

    ################################################################################################
    # VALUE serializers
    ################################################################################################

    def serialize_unit(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'unit', Empty()

    def serialize_bool(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'bool', to_boolean(obj)

    def serialize_text(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'text', to_str(obj)

    def serialize_int(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'int64', to_int(obj)

    def serialize_decimal(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'decimal', str(to_decimal(obj))

    def serialize_party(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'party', to_str(obj)

    def serialize_date(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'date', pb_get_date(to_date(obj))

    def serialize_datetime(self, context: TypeEvaluationContext, obj: Any) -> R:
        return 'timestamp', pb_get_timestamp(to_datetime(obj))

    def serialize_timedelta(self, context: TypeEvaluationContext, obj: Any) -> R:
        raise ValueError('RelTime types are not supported by the gRPC API')

    def serialize_contract_id(
            self, context: TypeEvaluationContext, tt: ContractIdType, obj: Any) -> R:
        return 'contract_id', to_str(obj)

    def serialize_optional(self, context: TypeEvaluationContext, tt: OptionalType, obj: Any):
        ut = tt.type_parameter
        if tt.internal_type is not None:
            blown_out_rep = {'None': {}} if obj is None else {'Some': obj}
            return self._serialize_dispatch(context.append_path('?'), tt.internal_type, blown_out_rep)
        else:
            from ..._gen.com.digitalasset.ledger.api.v1.value_pb2 import Optional, Value
            optional_message = Optional()
            if obj is not None:
                ctor, val = self._serialize_dispatch(context.append_path('?'), ut, obj)
                _set_value(optional_message.value, ctor, val)
            return 'optional', optional_message

    def serialize_list(self, context: TypeEvaluationContext, tt: ListType, obj: Any) -> R:
        from ..._gen.com.digitalasset.ledger.api.v1.value_pb2 import List
        ut = tt.type_parameter

        list_message = List()
        for i, item in enumerate(obj):
            value = list_message.elements.add()
            ctor, val = self._serialize_dispatch(context.append_path(f'[{i}]'), ut, item)
            _set_value(value, ctor, val)
        return 'list', list_message

    def serialize_map(self, context: TypeEvaluationContext, tt: MapType, obj: Any) -> R:
        if tt.internal_type is not None:
            blown_out_rep = {'Map_internal': [{'_1': k, '_2': v} for k, v in obj.items()]}
            return self._serialize_dispatch(context, tt.internal_type, blown_out_rep)
        else:
            raise ValueError('A native Map type does not yet exist in the gRPC API')

    def serialize_record(self, context: TypeEvaluationContext, tt: RecordType, obj: Any) -> R:
        from ..._gen.com.digitalasset.ledger.api.v1.value_pb2 import Record

        did_fail = False
        record_message = Record()
        for key, vt in tt.named_args:
            if key not in obj:
                LOG.error('The field %s is missing on %s', key, tt)
                did_fail = True
                continue

            ctor, value = self._serialize_dispatch(context.append_path(key), vt, obj.get(key))
            field = record_message.fields.add()
            field.label = key
            _set_value(field.value, ctor, value)
        if did_fail:
            raise ValueError('Failed to parse a record; check the logs for more information.')
        return 'record', record_message

    def serialize_variant(self, context: TypeEvaluationContext, tt: VariantType, obj: Any) -> R:
        from ..._gen.com.digitalasset.ledger.api.v1.value_pb2 import Variant
        try:
            obj_ctor, obj_value = decode_variant_dict(obj)
        except ValueError:
            if len(tt.named_args) == 1:
                # there is only one variation on this variant; under very specific circumstances
                # we'll allow for some convenience representations of this case
                obj_ctor, vt = tt.named_args[0]
                if (isinstance(vt, (RecordType, VariantType)) and len(vt.named_args) == 0) or \
                    vt == SCALAR_TYPE_UNIT:
                    obj_value = {}
                else:
                    LOG.error('Could not find a helpful representation of the single-variant case with value type %s', vt)
                    raise
            else:
                raise

        vt = tt._find_ctor(obj_ctor)

        ctor, value = self._serialize_dispatch(context.append_path(obj_ctor), vt, obj_value)

        variant_message = Variant()
        variant_message.constructor = obj_ctor
        _set_value(variant_message.value, ctor, value)
        return 'variant', variant_message

    def serialize_unsupported(
            self, context: TypeEvaluationContext, tt: UnsupportedType, obj: Any) -> R:
        raise Exception(f'UnsupportedType {tt} is not serializable in gRPC')


def _set_value(message: G.Value, ctor, value) -> None:
    """
    Work around the somewhat crazy API of Python's gRPC library to apply a known value to a
    :class:`Value`.

    :param message: The :class:`Value` object to modify.
    :param ctor: The actual field to apply to.
    :param value: The actual value to set. Must be compatible with the appropriate field.
    """
    try:
        if ctor == 'unit':
            message.unit.SetInParent()
        elif ctor in ('record', 'variant', 'list', 'optional'):
            getattr(message, ctor).MergeFrom(value)
        else:
            setattr(message, ctor, value)
    except:
        LOG.error('Failed to set a value %s, %s', ctor, value)
        raise



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains the description of the Ledger Query API's abstract syntax tree and
convenience methods for walking the tree.
"""

from .api import from_, init
from .builtins import BUILTIN_AND
from .expr_impl import and_, AppExpr, TemplateExpr
from .appbuilder import Application, ApplicationBuilder
from .appregister import register_application




# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains configuration parameters used by the rest of the library.
"""

import warnings
from argparse import Namespace

from dataclasses import dataclass, field, fields, asdict
from typing import Any, Collection, List, Mapping, Optional, Sequence

from .. import LOG
from ..model.core import ConfigurationError, Party
from ..util.config_meta import config_field, \
    COUNT_TYPE, LOG_LEVEL_TYPE, PARTIES_TYPE, PATH_TYPE, PORT_TYPE, SECONDS_TYPE, STRING_TYPE, \
    URL_TYPE, VERIFY_SSL_TYPE, config_fields, add_argument

# If this environment variable is set, is used in place of a configuration file if none is supplied
# on the command-line.
DAZL_CONFIG_ENV = 'DAZL_CONFIG'


@dataclass(frozen=True)
class _PartyConfig:
    """
    Configuration for a specific party's client.
    """

    url: str = config_field(
        'URL where the Ledger API is hosted',
        param_type=URL_TYPE,
        short_alias='u',
        environment_variable='DAML_LEDGER_URL',
        deprecated_alias='participant_url')

    admin_url: Optional[str] = config_field(
        'URL where administrative functions are exposed',
        param_type=URL_TYPE,
        environment_variable='DAML_LEDGER_ADMIN_URL',
        default_value=None)

    ca_file: Optional[str] = config_field(
        'server certificate authority file',
        param_type=PATH_TYPE,
        default_value=None)

    cert_file: Optional[str] = config_field(
        'client certificate file',
        param_type=PATH_TYPE)

    cert_key_file: Optional[str] = config_field(
        'client certificate and key file',
        param_type=PATH_TYPE)

    verify_ssl: Optional[str] = config_field(
        'level of verification to use for SSL',
        param_type=VERIFY_SSL_TYPE)

    poll_interval: Optional[float] = config_field(
        'polling interval for receiving new events',
        param_type=SECONDS_TYPE,
        default_value=0.1)

    connect_timeout: Optional[float] = config_field(
        'number of seconds before giving up on a connection',
        param_type=SECONDS_TYPE,
        default_value=120)

    application_name: Optional[str] = config_field(
        'The name that this application uses to identify itself to the ledger.',
        param_type=STRING_TYPE,
        default_value='DAZL-Client')

    max_event_block_size: Optional[int] = config_field(
        'Maximum number of blocks to read in a single call.',
        param_type=COUNT_TYPE,
        default_value=100)

    max_command_batch_size: Optional[int] = config_field(
        'Maximum number of commands to batch up in a single transaction',
        param_type=COUNT_TYPE,
        default_value=100)

    max_command_batch_timeout: Optional[float] = config_field(
        'Maximum number of seconds to wait before sending out a command',
        param_type=SECONDS_TYPE,
        default_value=0.25)

    party_groups: Optional[Collection[str]] = config_field(
        'comma-separated list of broadcast parties',
        param_type=PARTIES_TYPE)


@dataclass(frozen=True)
class _NetworkConfig:
    """
    Configuration for the entire client library and all connected parties.
    """

    log_level: Optional[int] = config_field(
        'logging level for events for this party',
        param_type=LOG_LEVEL_TYPE,
        environment_variable='DAZL_LOG_LEVEL',
        short_alias='l')

    max_connection_count: Optional[int] = config_field(
        'Number of concurrent HTTP connections to have outstanding.',
        param_type=COUNT_TYPE,
        default_value=20)

    quiet_timeout: Optional[float] = config_field(
        'Number of seconds to wait after the client "thinks" it\'s done to hang around for',
        param_type=SECONDS_TYPE,
        default_value=1)

    idle_timeout: Optional[float] = config_field(
        'Maximum number of seconds of idle activity before automatically closing the client',
        param_type=SECONDS_TYPE,
        default_value=120)

    max_consequence_depth: Optional[int] = config_field(
        'The maximum number of times to wait for all parties to arrive at the same offset' +
        'before failing with an error',
        param_type=COUNT_TYPE,
        default_value=50)

    server_host: Optional[str] = config_field(
        'Server listening host. Used for OAuth web application flow callbacks.',
        param_type=STRING_TYPE)

    server_port: Optional[int] = config_field(
        'Server listening port. Used for OAuth web application flow callbacks.',
        param_type=PORT_TYPE)

    oauth_client_id: Optional[str] = config_field(
        'OAuth client ID',
        param_type=STRING_TYPE)

    oauth_client_secret: Optional[str] = config_field(
        'OAuth client secret (implies web application flow or backend application flow)',
        param_type=STRING_TYPE)

    oauth_token: Optional[str] = config_field(
        'OAuth token',
        param_type=STRING_TYPE)

    oauth_refresh_token: Optional[str] = config_field(
        'OAuth refresh token',
        param_type=STRING_TYPE)

    oauth_id_token: Optional[str] = config_field(
        'OpenID token (JWT formatted)',
        param_type=STRING_TYPE)

    oauth_token_uri: Optional[str] = config_field(
        'OAuth token URL',
        param_type=STRING_TYPE)

    oauth_redirect_uri: Optional[str] = config_field(
        'OAuth redirect URI (implies web application flow)',
        param_type=STRING_TYPE)

    oauth_auth_url: Optional[str] = config_field(
        'OAuth auth URL (implies mobile application flow)',
        param_type=STRING_TYPE)

    oauth_legacy_username: Optional[str] = config_field(
        'OAuth username (implies legacy application flow)',
        param_type=STRING_TYPE)

    oauth_legacy_password: Optional[str] = config_field(
        'OAuth password (implies legacy application flow)',
        param_type=STRING_TYPE)


@dataclass(frozen=True)
class _FlatConfig(_NetworkConfig, _PartyConfig):
    """
    A configuration object that assumes a single flat key-value structure suitable for parsing
    command-line or environment variable-based values.

    This configuration object is used as an intermediate holding object when parsing parameters
    from environment variables or command-line arguments.
    """
    parties: Collection[Party] = config_field(
        'comma-separated list of parties serviced by a participant node',
        param_type=PARTIES_TYPE,
        short_alias='p',
        environment_variable='DAML_LEDGER_PARTY',
        default_value=())


@dataclass(frozen=True)
class PartyConfig(_PartyConfig):
    party: Optional[Party] = None


@dataclass(frozen=True)
class NetworkConfig(_NetworkConfig):
    parties: 'Sequence[PartyConfig]' = field(default_factory=tuple)


def configure_parser(arg_parser, config_file_support=False):
    """
    Add standard options to an arg parser (later to be extracted out by ``get_config``).
    """
    if config_file_support:
        from os import environ
        default_config = environ.get(DAZL_CONFIG_ENV)
        group = arg_parser.add_argument_group('Configuration Settings')
        group.add_argument('--config', help='path to a YAML config file', default=default_config)

    party_config_fields = [fld.name for fld, _ in config_fields(_PartyConfig)]

    group = arg_parser.add_argument_group('Overall Settings')
    for fld, param in config_fields(_FlatConfig):
        if fld.name not in party_config_fields:
            add_argument(group, fld.name, param)

    group = arg_parser.add_argument_group('Per-Party Settings')
    for fld, param in config_fields(_PartyConfig):
        add_argument(group, fld.name, param)

    return arg_parser


def unflatten_config(config: '_FlatConfig') -> 'NetworkConfig':
    """
    Convert a :class:`_FlatConfig` to a :class:`NetworkConfig`.
    """
    config_dict = {k: v for k, v in asdict(config).items() if v is not None}
    network_dict = {fld.name: config_dict.get(fld.name) for fld in fields(_NetworkConfig)}
    party_dict = {fld.name: config_dict.get(fld.name) for fld in fields(_PartyConfig)}
    parties = tuple(PartyConfig(party=party, **party_dict) for party in config.parties)
    return NetworkConfig(parties=parties, **network_dict)


def parse_kwargs(*config: 'NetworkConfig', **kwargs) -> 'NetworkConfig':
    """
    Create a :class:`NetworkConfig` for the configuration settings.

    :param config:
        Instances of :class:`NetworkConfig`` objects to merge together.
    :param kwargs:
        Configuration options that are accepted either at the global level or at the party level.
    :return:
        An instance of :class:`NetworkConfig`.
    """
    configurations = list(config)  # type: List[NetworkConfig]
    if kwargs:
        args_dict = dict(_get_env_defaults())
        args_dict.update(_parse_args_dict(kwargs))
        flat_config = _FlatConfig(**args_dict)
        configurations.append(unflatten_config(flat_config))
    network_config = merge_configurations(*configurations)
    return validate_config(network_config)


def get_config(args: 'Namespace', config_file_support: bool = False) -> 'NetworkConfig':
    """
    Convert an ``argparse.Namespace`` to a fully-formed and valid :class:`NetworkConfig`.
    """
    args_dict = dict(_get_env_defaults())
    args_dict.update(_parse_args_ns(args))
    flat_config = _FlatConfig(**args_dict)
    network_config = unflatten_config(flat_config)
    return validate_config(network_config)


def fetch_config(path_or_url: Optional[str]) -> Optional[str]:
    """
    Attempts to fetch a config file from what looks like either a path or a URL.

    :param path_or_url: A file path or a URL.
    :return: The body of the config file.
    """
    if not path_or_url:
        return None

    from urllib.error import URLError, HTTPError
    from urllib.parse import urlparse
    from urllib.request import urlopen

    u = urlparse(path_or_url)
    if not u.scheme or u.scheme == 'file':
        try:
            with open(u.path, 'r') as buf:
                config_str = buf.read()
        except FileNotFoundError:
            raise ConfigurationError([f'Config file not found: {u.path}'])
    else:
        try:
            with urlopen(path_or_url) as buf:
                config_str = buf.read().decode('utf8')
        except HTTPError as error:
            raise ConfigurationError([f'HTTP {error.code}: {path_or_url}'])
        except URLError as error:
            if isinstance(error.args[0], ConnectionRefusedError):
                raise ConfigurationError([f'HTTP unreachable: {path_or_url}'])
            else:
                raise ConfigurationError([error.args[0]])

    return config_str


def _get_env_defaults() -> 'Mapping[str, Any]':
    from os import getenv
    kwargs = {}
    for fld, param in config_fields(_FlatConfig):
        if param.environment_variable:
            value = getenv(param.environment_variable)
            if value:
                kwargs[fld.name] = param.param_type.value_ctor(value)
    return kwargs


def _parse_args_ns(args: 'Namespace') -> 'Mapping[str, Any]':
    """
    Convert the values
    :param args:
    :return:
    """
    return {fld.name: getattr(args, fld.name)
            for fld in fields(_FlatConfig) if hasattr(args, fld.name)}


def _parse_args_dict(d: 'Mapping[str, Any]') -> 'Mapping[str, Any]':
    """
    Convert the key-value pairs in this mapping to keys that are defined on :class:`_FlatConfig`.

    :param d:
    :return:
    """
    kwargs = {}
    unknown_fields = set()
    for fld, param in config_fields(_FlatConfig):
        value = d.get(fld.name)
        if value is None:
            for key in param.alternate_keys:
                if d.get(key) is not None:
                    value = d.get(key)
                    break

        if value is None:
            for key in param.deprecated_aliases:
                if d.get(key) is not None:
                    warnings.warn(f'The {key} option is deprecated. Please use {fld.name} instead.',
                                  DeprecationWarning)
                    value = d.get(key)
                    break

        if value is not None:
            if param.param_type.value_ctor is not None:
                try:
                    value = param.param_type.value_ctor(value)
                except:
                    LOG.exception('Config key: %r; could not parse value: %r', fld.name, value)

            kwargs[fld.name] = value

    if unknown_fields:
        raise ValueError(f'unknown kwargs: {sorted(unknown_fields)}')
    return kwargs


def validate_config(config: 'NetworkConfig') -> 'NetworkConfig':
    failures = set()

    LOG.debug('Configuration: %s', config)

    if config.parties:
        for party_config in config.parties:
            if not party_config.url:
                failures.add(f'Party {party_config.party} has no Ledger API URL')
    else:
        failures.add('At least one party and a URL must be specified')

    if failures:
        raise ConfigurationError(failures)

    return config


def merge_configurations(*configurations: 'NetworkConfig') -> 'NetworkConfig':
    return configurations[0]



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from io import BytesIO
from os import path
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import BinaryIO, Dict, List, Mapping, Optional, Union, TYPE_CHECKING, Sequence
from zipfile import ZipFile

from .process import ProcessWatcher
from ..util.path_util import pathify

if TYPE_CHECKING:
    from ..model.types_store import PackageStore, PackageProvider


def build_dar(
        daml_path: 'Union[str, Path]',
        dar_path: str,
        damlc_component: 'Optional[str]' = None,
        damlc_extra_args: 'Sequence[str]' = None) -> None:
    """
    Create a .dar file from a .daml file.

    :param daml_path: Path to the DAML file.
    :param dar_path: Path to the DAR file to create.
    :param damlc_component: Version of the DAML compiler to use.
    :param damlc_extra_args: Extra arguments of damlc.
    """
    from ..damlsdk.package import package, PackageOptions
    options = PackageOptions([str(daml_path)], dar_path, damlc_extra_args)
    proc_opts = package(options, component=damlc_component)
    with ProcessWatcher(proc_opts) as pw:
        code = pw.run()
    if code != 0:
        raise DamlcPackageError(code)


def get_archives(contents: bytes) -> 'Mapping[str, bytes]':
    """
    Attempt to parse the specified contents as either a DALF or a DAR, and return a mapping of
    package IDs to DALF bytes.

    :param contents: A byte array to attempt to parse.
    :return: Return a mapping from package ID to byte contents.
    """
    if contents[0:4] == b'PK\03\04':
        with BytesIO(contents) as buf:
            with DarFile(buf) as dar:
                return dar.get_archives()
    else:
        # parse as a single DALF
        from ..protocols.v1.pb_parse_metadata import parse_archive_payload
        a = parse_archive_payload(contents)
        return {a.hash: contents}


class TemporaryDar:
    """
    Wrapper that creates a .dar file from .daml files that is only present on the file system
    during the scope of this object.
    """
    __slots__ = ('daml_path', 'dar_paths', 'damlc_component', '_tmp_dir', 'damlc_extra_args')

    daml_path: str
    dar_paths: Optional[List[str]]
    damlc_component: Optional[str]
    _tmp_dir: TemporaryDirectory
    damlc_extra_args: Sequence[str]

    def __init__(self, daml_path: Union[str, Path], damlc_component: Optional[str] = None, damlc_extra_args=None):
        self.daml_path = pathify(daml_path)
        self.dar_paths = None
        self.damlc_component = damlc_component
        self._tmp_dir = TemporaryDirectory()
        self.damlc_extra_args = damlc_extra_args

    def store(self) -> 'PackageStore':
        from .dar_repo import LocalDarRepository
        repo = LocalDarRepository()
        repo.add_source(*self.ensure_dar())
        return repo.store

    def ensure_dar(self) -> List[str]:
        if self.dar_paths is None:
            dar_path = path.join(self._tmp_dir.name, 'file.dar')
            build_dar(self.daml_path, dar_path, damlc_component=self.damlc_component,
                      damlc_extra_args=self.damlc_extra_args)
            from os import listdir
            self.dar_paths = [path.join(self._tmp_dir.name, o) for o in listdir(self._tmp_dir.name)]

        return self.dar_paths

    def __enter__(self) -> List[str]:
        return self.ensure_dar()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.cleanup()

    def cleanup(self) -> None:
        if self._tmp_dir is not None:
            self._tmp_dir.cleanup()
            self._tmp_dir = None


class DarFile:
    """
    Provides access to the contents of a .dar file.
    """
    def __init__(self, dar_path: 'Union[str, Path, BinaryIO]'):
        if isinstance(dar_path, (str, Path)):
            self.dar_path = pathify(dar_path)
            self.dar_contents = ZipFile(str(self.dar_path))
        else:
            self.dar_path = None
            self.dar_contents = ZipFile(dar_path)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dar_contents.close()

    def read_metadata(self) -> 'PackageStore':
        from ..model.types_store import PackageStore
        store = PackageStore.empty()
        dalf_names = self.get_dalf_names()
        for dalf_name in dalf_names:
            contents = self.dar_contents.read(dalf_name)
            store.register_all(parse_dalf(contents))
        return store

    def get_archives(self) -> 'Mapping[str, bytes]':
        """
        Return a mapping from package ID to byte contents.
        """
        from ..protocols.v1.pb_parse_metadata import parse_archive_payload
        archives = {}  # type: Dict[str, bytes]
        for dalf_name in self.get_dalf_names():
            contents = self.dar_contents.read(dalf_name)
            payload = parse_archive_payload(contents)
            archives[payload.hash] = contents
        return archives

    def get_dalf_names(self) -> 'Sequence[str]':
        dalf_names = []
        for name in self.dar_contents.namelist():
            _, ext = path.splitext(name)
            if ext == '.dalf':
                dalf_names.append(name)
        return dalf_names

    def get_manifest(self) -> 'Optional[Mapping[str, str]]':
        """
        Return the contents of the manifest of this DAR.
        :return:
        """
        names = self.dar_contents.namelist()
        if 'META-INF/MANIFEST.MF' in names:
            manifest_bytes = self.dar_contents.read('META-INF/MANIFEST.MF')
            manifest = {}
            for line in manifest_bytes.decode('utf-8').splitlines():
                print(line)
                name, _, value = line.partition(':')
                manifest[name] = value.strip()
            return manifest
        else:
            return None

    def get_sdk_version(self) -> 'Optional[str]':
        """
        Return the SDK version used to compile this dar (if this information is available).
        """
        manifest = self.get_manifest()
        return manifest.get('Sdk-Version') if manifest is not None else None

    def get_package_provider(self) -> 'PackageProvider':
        from typing import Dict
        from ..model.types_store import MemoryPackageProvider
        from .._gen.da.daml_lf_pb2 import Archive

        packages = {}  # type: Dict[str, bytes]
        dalf_names = self.get_dalf_names()
        for dalf_name in dalf_names:
            contents = self.dar_contents.read(dalf_name)

            a = Archive()
            a.ParseFromString(contents)

            packages[a.hash] = a.payload

        return MemoryPackageProvider(packages)


def parse_dalf(contents: bytes) -> 'PackageStore':
    from .._gen.da.daml_lf_pb2 import Archive
    from ..protocols.v1.pb_parse_metadata import parse_archive_payload, parse_daml_metadata_pb
    a = Archive()
    a.ParseFromString(contents)
    p = parse_archive_payload(a.payload)
    return parse_daml_metadata_pb(a.hash, p)


class DamlcPackageError(Exception):
    """
    Raised when DAMLC fails to compile a DAR.
    """
    def __init__(self, exit_code):
        self.exit_code = exit_code



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module implements a rewriter over arbitrary ``tuple`` instances.
"""

from collections import defaultdict
from typing import Any, Callable, Tuple


from .. import LOG


class RewriteContext:
    """
    Recursive object rewriter, based on type-specific callbacks.

    For tuple types, constituent values are recursively rewritten, and if any value is actually
    changed, a new tuple of the same type is created.
    """
    def __init__(self):
        self._type_dispatch = dict()
        self._context_objects = defaultdict(set)

    def context_object(self, typ, name) -> None:
        """
        Supply objects of the specified type as named parameters to rewrite callbacks. Because
        rewriting happens bottom-up, these context objects will necessarily be pre-modified
        (original) values. Additionally, the context object will only be supplied on *child*
        objects, and not the original handler.

        :param typ:
        :param name:
        """
        self._context_objects[typ].add(name)

    def register(self, typ, fn=None, *, recurse=None):
        if fn is None:
            return lambda func: self.register(typ, func, recurse=recurse)
        else:
            self._type_dispatch[typ] = (fn, recurse)
            return fn

    def register_ro(self, typ, fn=None, *, recurse=None):
        """
        Register a handler that does not modify the visited node (the return value is ignored).
        """
        def _wrapper(value, *args, **kwargs):
            fn(value, *args, **kwargs)
            return value
        self.register(typ, fn=_wrapper, recurse=recurse)
        return fn

    def __call__(self, value, *args, **kwargs):
        return self.rewrite(value, *args, **kwargs)

    def rewrite(self, tuple_value, *args, **kwargs):
        # ``None`` cannot be rewritten
        if tuple_value is None:
            return None

        tuple_class = type(tuple_value)
        fn, recurse = self.dispatch(tuple_class)
        if recurse is None:
            recurse = issubclass(tuple_class, (tuple, list, dict)) or \
                      (hasattr(tuple_value, '__getnewargs__') and hasattr(tuple_value, '__dict__'))

        if recurse:
            if tuple_class == tuple:
                # Tuples are frequently subclassed, with different __getnewargs__ behavior for most subtypes.
                # Assume that raw tuples are to be unpacked, and anything else is treated specially (possibly
                # caught by other branches, particularly __getnewargs__).
                old_values = tuple_value
                ctor = lambda *values: tuple(values)
            elif hasattr(tuple_value, '__getnewargs__'):
                old_values = tuple_value.__getnewargs__()
                ctor = tuple_class
            elif isinstance(tuple_value, dict):
                old_values = tuple(tuple_value.values())
                ctor = lambda *values: dict(zip(tuple_value.keys(), values))
            elif isinstance(tuple_value, list):
                old_values = tuple_value
                ctor = lambda *values: list(values)
            else:
                old_values = tuple_value
                ctor = tuple_class

            child_kwargs = dict(kwargs)
            extra_kwargs = {n: tuple_value for n in self._context_objects.get(tuple_class, ())}
            child_kwargs.update(extra_kwargs)

            new_values = [self.rewrite(i, *args, **child_kwargs) for i in old_values]
            changed = False
            for old_value, new_value in zip(old_values, new_values):
                if old_value is not new_value:
                    changed = True
                    break
            if changed:
                try:
                    tuple_value = ctor(*new_values)
                except:
                    LOG.error('Failed to construct %r with arguments %r', ctor, new_values)
                    raise

        return fn(tuple_value, *args, **kwargs)

    def dispatch(self, typ) -> Tuple[Callable[..., Any], bool]:
        disp = self._type_dispatch.get(typ)
        return disp if disp is not None else (_identity, None)


# noinspection PyUnusedLocal
def _identity(value, *args, **kwargs):
    """
    Merely return the original value, unchanged.

    :param value: The value to return.
    :param args: Unused.
    :param kwargs: Unused.
    :return: The original value.
    """
    return value



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains builders for applications.
"""
import collections

from ._rewriter import RewriteContext
from .expr_impl import SuiteExpr, ProgramExpr, TemplateSelectExpr, Application


class ApplicationBuilder:
    """
    Builder for an application.
    """
    def __init__(self):
        self.suite_builders = collections.OrderedDict()

    def as_party(self, party_name):
        """
        Creates a builder for a `SuiteExpr` to be run under the context of the named party.
        """
        suite_builder = self.suite_builders.get(party_name)
        if suite_builder is None:
            suite_builder = SuiteBuilder(party_name)
            self.suite_builders[party_name] = suite_builder
        return suite_builder

    def build(self) -> Application:
        """
        Builds an application.
        """
        return Application([builder.build() for builder in self.suite_builders.values()])


class SuiteBuilder:
    """
    Builder for a suite of programs.
    """
    def __init__(self, party_name):
        self.party_name = party_name
        self.programs = []

    def add(self, program):
        """
        Registers a program to watch and respond to changes in state on the ledger.

        :param program:
            A `ProgramExpr` that defines a search condition and an operation to perform against
            all members of the set that match the search condition.
        :return:
            `self`; this class is intended to be used in a fluent style.
        """
        self.programs.append(generate_default_aliases(ProgramExpr.coerce(program)))
        return self

    def extend(self, programs):
        self.programs.extend(map(generate_default_aliases, programs))
        return self

    def build(self):
        """
        Build a `SuiteExpr` that are the instructions to perform on the ledger.
        """
        return SuiteExpr(party=self.party_name, programs=list(self.programs))


def merge_suites(*suites):
    """
    Merges multiple suites together.
    """
    suites = [suite for suite in suites if suite]
    if not suites:
        raise ValueError("at least one suite required here")

    # TODO: The parties must all be the same; verify this
    party = suites[0].party
    programs = []
    for suite in suites:
        programs.extend(suite.programs)
    return SuiteExpr(party=party, programs=programs)


def generate_default_aliases(expr):
    template_refs = set()

    context = RewriteContext()

    @context.register(TemplateSelectExpr)
    def capture_distinct_selects(select):
        template_refs.add(_named_template_ref(select))
        return select

    context.rewrite(expr)

    unnamed_refs = {tref.template_id for tref in template_refs if tref.alias is None}
    used_ref_names = {tref.alias for tref in template_refs if tref.alias is not None}

    rewritten_template_refs = {
        NamedTemplateRef(template_id=tref, alias=None):
        NamedTemplateRef(template_id=tref, alias=generate_alias_name(used_ref_names, tref.name, i))
        for i, tref in enumerate(unnamed_refs)}

    @context.register(TemplateSelectExpr)
    def rewrite_select(select: TemplateSelectExpr):
        rewrite = rewritten_template_refs[_named_template_ref(select)]
        return TemplateSelectExpr(template=rewrite.template_id, alias=rewrite.alias)

    return context.rewrite(expr)


def generate_alias_name(used_names, template_id: str, i: int):
    candidate = template_id.split('.')[-1][0].lower() + str(i)
    return candidate if candidate not in used_names else candidate + '0'


def _named_template_ref(expr: TemplateSelectExpr):
    return NamedTemplateRef(expr.template.template_ref, expr.alias)


NamedTemplateRef = collections.namedtuple('NamedTemplateRef', ['template_id', 'alias'])



# Copyright (c) 2019 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains utilities for starting up a local instance of the Ledger Sandbox,
either through the Sandbox or one local to the build tree of the project.
"""

import logging
from contextlib import ExitStack
from pathlib import Path
from typing import Collection, Sequence, Optional, Type, Union, Tuple, ContextManager

from dataclasses import dataclass

from .. import LOG
from .fetch import sdk_component_path
from ..model.core import DazlError
from ..model.ledger import TimeModel, StaticTimeModel, RealTimeModel
from ..model.network import SSLSettings
from ..util.process import ProcessContext, ProcessWatcher


@dataclass(frozen=True)
class SandboxOptions:
    """
    Parameters that indicate how a sandbox is to be started.
    """
    files: 'Sequence[Union[str, Path]]'
    port: 'int'
    extra_args: 'Sequence[str]'


def sandbox(daml_path: 'Union[str, Path, Collection[str], Collection[Path]]',
            port: 'Optional[int]' = None,
            backend: 'Optional[str]' = None,
            damlc: 'Optional[str]' = None,
            time_model_type: 'Type[TimeModel]' = StaticTimeModel,
            ssl_settings: 'Optional[SSLSettings]' = None,
            extra_args: 'Optional[Sequence[str]]' = None,
            damlc_extra_args: 'Optional[Sequence[str]]' = None) -> ProcessWatcher:
    """
    Run an instance of the Sandbox.

    :param daml_path:
        The path to the DAML file or DAR file to open, or a list of files.
    :param port:
        The port to bind the server to, or ``None`` to pick a port at random and use it.
    :param backend:
        ``None`` to use the default backend (``'sdk'`` if it is found, otherwise ``'damlc'``) or a
        string that identifies a backend.
    :param damlc:
        ``None`` to use the default DAML compiler (``'sdk'`` if it found) or a string that
        identifies a DAML compiler.
    :param time_model_type:
        The mode in which to run the Sandbox clock. Defaults to static time.
    :param ssl_settings:
        Settings that configure how the Sandbox is to start in HTTPS mode. If not supplied, the
        Sandbox starts up in HTTP.
    :param extra_args:
        Extra arguments to pass to the Sandbox.
    :param damlc_extra_args:
        Extra arguments to pass to the compiler (only used when compilation is performed as part
        of running the Sandbox).
    :return:
        A ``SandboxRunner`` that can be used as a context object to end the process.
    """
    # provide default values for backend and port if they were not provided
    if daml_path is None:
        raise ValueError('daml_path is required')

    if port is None:
        from ..util.io import find_free_port
        port = find_free_port()

    full_extra_args = [] if not extra_args else list(extra_args)
    if time_model_type == StaticTimeModel:
        pass
    elif time_model_type == RealTimeModel:
        full_extra_args.append('-w')
    if ssl_settings:
        if ssl_settings.cert_file:
            full_extra_args.extend(('--crt', ssl_settings.cert_file))
        if ssl_settings.cert_key_file:
            full_extra_args.extend(('--pem', ssl_settings.cert_key_file))
        if ssl_settings.ca_file:
            full_extra_args.extend(('--cacrt', ssl_settings.ca_file))

    if isinstance(daml_path, str):
        daml_path = [daml_path]
    elif isinstance(daml_path, Path):
        daml_path = [str(daml_path)]

    options = SandboxOptions(files=daml_path, port=port, extra_args=full_extra_args)
    proc_opts = _sandbox(options, backend, damlc_component=damlc, damlc_extra_args=damlc_extra_args)
    pw = ProcessWatcher(proc_opts)
    pw.extra_data = proc_opts.extra_data
    pw.url = f'http://localhost:{port}'
    return pw


@dataclass(frozen=True)
class _SandboxIntermediateOptions:
    args: 'Tuple[str]'
    context: 'Optional[ContextManager]'
    files: 'Collection[str]'


def _sandbox(options: SandboxOptions,
             component: 'Optional[str]' = None,
             damlc_component: 'Optional[str]' = None,
             damlc_extra_args: 'Sequence[str]' = None) \
        -> 'ProcessContext':
    name, path = sdk_component_path(component or 'sandbox')
    if name == 'damlc':
        # older versions of the DAML compiler supported running DAML natively, but this is no
        # longer supported
        raise ValueError('The damlc-based sandbox is no longer supported.')
    elif name == 'sdk':
        # older versions of dazl could use 'da run sandbox' directly, but the DA Assistant will
        # soon be deprecated and equivalent functionality will not be replicated
        raise ValueError('SDK-based running of the sandbox is no longer supported.')
    elif name != 'sandbox':
        # the sandbox is really the only supported backend at this point
        raise ValueError(f'Unknown component type: {component}')

    LOG.debug('Starting sandbox %s...', component)

    # start the JVM-based sandbox, possibly compiling the source files if they are not already
    # .dar packages or .dalf archives
    int_opts = _sandbox_options(options, damlc_component, damlc_extra_args)
    sandbox_jar = next(path.glob('sandbox-*.jar'))

    args = ['java', '-jar', sandbox_jar, *int_opts.args]
    return ProcessContext(
        args, secondary_context=int_opts.context, logger=logging.getLogger('sandbox'),
        watch_port=options.port, extra_data={'files': int_opts.files})


def _sandbox_options(options: 'SandboxOptions',
                     damlc_component: 'Optional[str]',
                     damlc_extra_args: 'Sequence[str]') \
        -> '_SandboxIntermediateOptions':
    from ..util.dar import TemporaryDar

    files = []
    context = ExitStack()
    for file in options.files:
        ext = Path(file).suffix.lower()
        if ext == '.daml':
            # the Java Sandbox needs .daml files to be converted to a package first
            temp_dar = TemporaryDar(
                file, damlc_component=damlc_component, damlc_extra_args=damlc_extra_args)
            files.extend(context.enter_context(temp_dar))
        elif ext == '.dalf':
            files.append(file)
        elif ext == '.dar':
            files.append(file)
        else:
            raise ValueError(f'unknown file type: {ext}')

    return _SandboxIntermediateOptions(
        args=tuple(str(s) for s in [*options.extra_args, '--port', options.port, *files]),
        context=context,
        files=files)


class SandboxError(DazlError):
    """
    Error raised when the Sandbox aborts abnormally.
    """
    def __init__(self, exit_code: int):
        self.exit_code = exit_code
